# Description: Schema Document
from dataclasses import dataclass

from scaledp.schemas.Box import Box
from scaledp.utils.dataclass import map_dataclass_to_struct, register_type


@dataclass(order=True)
class Document:
    path: str
    text: str
    type: str
    bboxes: list[Box]
    exception: str = ""

    @staticmethod
    def get_schema():
        return map_dataclass_to_struct(Document)

    def merge(self, document):
        return Document(
            path=document.path,
            text=document.text + "\n" + self.text,
            type=document.type,
            bboxes=document.bboxes + self.bboxes,
            exception=document.exception,
        )


register_type(Document, Document.get_schema)



from typing import Any, Tuple

import cv2
import numpy as np
import onnxruntime

from scaledp.models.detectors.yolo.utils import multiclass_nms, xywh2xyxy


class YOLO:

    def __init__(self, path, conf_thres=0.7, iou_thres=0.5) -> None:
        self.conf_threshold = conf_thres
        self.iou_threshold = iou_thres

        # Store original image dimensions and scaling info
        self.original_width = None
        self.original_height = None
        self.scale_factor = None
        self.pad_x = None
        self.pad_y = None

        # Initialize model
        self.initialize_model(path)

    def __call__(self, image) -> Any:
        return self.detect_objects(image)

    def initialize_model(self, path):
        self.session = onnxruntime.InferenceSession(
            path, providers=onnxruntime.get_available_providers()
        )
        # Get model info
        self.get_input_details()
        self.get_output_details()

    def detect_objects(self, image):
        input_tensor = self.prepare_input(image)

        # Perform inference on the image
        outputs = self.inference(input_tensor)

        self.boxes, self.scores, self.class_ids = self.process_output(outputs)

        return self.boxes, self.scores, self.class_ids

    def rescale_image_with_padding(self, image: np.ndarray, target_size: Tuple[int, int]) -> np.ndarray:
        """
        Rescale image while keeping aspect ratio and pad with white background.

        Args:
            image: Input image (H, W, C)
            target_size: Target size (width, height)

        Returns:
            Rescaled and padded image
        """
        self.original_height, self.original_width = image.shape[:2]
        target_width, target_height = target_size

        # Calculate scaling factor to maintain aspect ratio
        scale_w = target_width / self.original_width
        scale_h = target_height / self.original_height
        self.scale_factor = min(scale_w, scale_h)

        # Calculate new dimensions
        new_width = int(self.original_width * self.scale_factor)
        new_height = int(self.original_height * self.scale_factor)

        # Resize image
        resized_image = cv2.resize(image, (new_width, new_height))

        # Calculate padding
        self.pad_x = (target_width - new_width) // 2
        self.pad_y = (target_height - new_height) // 2

        # Create padded image with white background
        padded_image = np.full((target_height, target_width, 3), 255, dtype=np.uint8)
        padded_image[self.pad_y:self.pad_y + new_height, self.pad_x:self.pad_x + new_width] = resized_image

        return padded_image

    def restore_coordinates(self, boxes: np.ndarray) -> np.ndarray:
        """
        Restore bounding box coordinates to original image space.

        Args:
            boxes: Bounding boxes in model input space (N, 4) [x1, y1, x2, y2]

        Returns:
            Bounding boxes in original image space
        """
        if len(boxes) == 0:
            return boxes

        restored_boxes = boxes.copy()

        # Remove padding offset
        restored_boxes[:, [0, 2]] -= self.pad_x  # x coordinates
        restored_boxes[:, [1, 3]] -= self.pad_y  # y coordinates

        # Scale back to original size
        restored_boxes = restored_boxes / self.scale_factor

        # Clip to original image bounds
        restored_boxes[:, [0, 2]] = np.clip(restored_boxes[:, [0, 2]], 0, self.original_width)
        restored_boxes[:, [1, 3]] = np.clip(restored_boxes[:, [1, 3]], 0, self.original_height)

        return restored_boxes

    def prepare_input(self, image):
        # Store original dimensions for coordinate restoration
        self.img_height, self.img_width = image.shape[:2]

        input_img = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        # Rescale image with padding instead of simple resize
        input_img = self.rescale_image_with_padding(input_img, (self.input_width, self.input_height))

        # Scale input pixel values to 0 to 1
        input_img = input_img / 255.0
        input_img = input_img.transpose(2, 0, 1)
        return input_img[np.newaxis, :, :, :].astype(np.float32)

    def inference(self, input_tensor):
        return self.session.run(self.output_names, {self.input_names[0]: input_tensor})

    def process_output(self, output):
        predictions = np.squeeze(output[0]).T

        # Filter out object confidence scores below threshold
        scores = np.max(predictions[:, 4:], axis=1)
        predictions = predictions[scores > self.conf_threshold, :]
        scores = scores[scores > self.conf_threshold]

        if len(scores) == 0:
            return [], [], []

        # Get the class with the highest confidence
        class_ids = np.argmax(predictions[:, 4:], axis=1)

        # Get bounding boxes for each object
        boxes = self.extract_boxes(predictions)

        # Apply non-maxima suppression to suppress weak, overlapping bounding boxes
        indices = multiclass_nms(boxes, scores, class_ids, self.iou_threshold)

        # Restore coordinates to original image space
        final_boxes = self.restore_coordinates(boxes[indices])

        return final_boxes, scores[indices], class_ids[indices]

    def extract_boxes(self, predictions):
        # Extract boxes from predictions
        boxes = predictions[:, :4]

        # Convert boxes to xyxy format (no rescaling yet, done in restore_coordinates)
        return xywh2xyxy(boxes)

    def rescale_boxes(self, boxes):

        # Rescale boxes to original image dimensions
        input_shape = np.array(
            [self.input_width, self.input_height, self.input_width, self.input_height]
        )
        boxes = np.divide(boxes, input_shape, dtype=np.float32)
        boxes *= np.array(
            [self.img_width, self.img_height, self.img_width, self.img_height]
        )
        return boxes

    def get_input_details(self):
        model_inputs = self.session.get_inputs()
        self.input_names = [model_inputs[i].name for i in range(len(model_inputs))]

        self.input_shape = model_inputs[0].shape
        self.input_height = (
            self.input_shape[2] if self.input_shape[2] != "height" else 960
        )
        self.input_width = (
            self.input_shape[3] if self.input_shape[3] != "width" else 960
        )

    def get_output_details(self):
        model_outputs = self.session.get_outputs()
        self.output_names = [model_outputs[i].name for i in range(len(model_outputs))]



import logging
import traceback
from types import MappingProxyType
from typing import Any

import fitz
from pyspark import keyword_only
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.pandas import DataFrame
from pyspark.sql.functions import udf
from pyspark.sql.types import Row

from scaledp.params import (
    HasColumnValidator,
    HasInputCol,
    HasKeepInputData,
    HasOutputCol,
    HasPageCol,
    HasPathCol,
    HasResolution,
)
from scaledp.schemas.Box import Box
from scaledp.schemas.Document import Document


class PdfDataToDocument(
    Transformer,
    HasInputCol,
    HasOutputCol,
    HasKeepInputData,
    HasPathCol,
    HasPageCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasColumnValidator,
    HasResolution,
):
    """Extract text with coordinates from PDF file."""

    DEFAULT_PARAMS = MappingProxyType(
        {
            "inputCol": "content",
            "outputCol": "document",
            "pathCol": "path",
            "pageCol": "page",
            "keepInputData": False,
            "resolution": 300,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(PdfDataToDocument, self).__init__()
        self._setDefault(**self.DEFAULT_PARAMS)
        self._set(**kwargs)

    def transform_udf(self, input: Row, path: str) -> list[Document]:
        logging.info("Run Pdf Data to Text")
        try:
            doc = fitz.open("pdf", input)
            if len(doc) == 0:
                raise ValueError("Empty PDF document.")

            page = doc[0]
            # Get the page's transformation matrix and dimensions
            dpi = self.getResolution()  # PDF default DPI

            # Get page rotation
            rotation = page.rotation
            # Normalize rotation to 0, 90, 180, or 270 degrees
            rotation = rotation % 360

            words = page.get_text("words")
            boxes = []
            text_content = []

            # Get page dimensions
            page_width = page.mediabox_size[0]
            page_height = page.mediabox_size[1]

            for word in words:
                x0, y0, x1, y1, word_text, _, _, _ = word
                # Convert PDF coordinates to pixel coordinates using the transformation matrix
                # Scale coordinates according to DPI
                scale = dpi / 72

                # Apply basic coordinate transformation
                if rotation == 0:
                    pixel_x0 = x0 * scale
                    pixel_y0 = y0 * scale
                    pixel_x1 = x1 * scale
                    pixel_y1 = y1 * scale
                elif rotation == 90:
                    # Rotate 90 degrees clockwise
                    pixel_x0 = page_height * scale - y0 * scale
                    pixel_y0 = x0 * scale
                    pixel_x1 = page_height * scale - y1 * scale
                    pixel_y1 = x1 * scale
                elif rotation == 180:
                    # Rotate 180 degrees
                    pixel_x0 = page_width * scale - x0 * scale
                    pixel_y0 = y0 * scale
                    pixel_x1 = page_width * scale - x1 * scale
                    pixel_y1 = y1 * scale
                elif rotation == 270:
                    # Rotate 270 degrees clockwise
                    pixel_x0 = y0 * scale
                    pixel_y0 = page_width * scale - x0 * scale
                    pixel_x1 = y1 * scale
                    pixel_y1 = page_width * scale - x1 * scale

                # Ensure coordinates are properly ordered (x0 < x1 and y0 < y1)
                x_min = min(pixel_x0, pixel_x1)
                x_max = max(pixel_x0, pixel_x1)
                y_min = min(pixel_y0, pixel_y1)
                y_max = max(pixel_y0, pixel_y1)

                boxes.append(
                    Box(
                        x=int(x_min),
                        y=int(y_min),
                        width=int(x_max - x_min),
                        height=int(y_max - y_min),
                        text=word_text,
                        score=1.0,
                    ),
                )
                text_content.append(word_text)

            return Document(
                path=path,
                text=" ".join(text_content),
                type="pdf",
                bboxes=boxes,
            )

        except Exception:
            exception = traceback.format_exc()
            exception = (
                f"{self.uid}: Error during extracting text from "
                f"the PDF document: {exception}"
            )
            logging.warning(exception)
            return [
                Document(
                    path=path,
                    text="",
                    type="pdf",
                    bboxes=[],
                    exception=exception,
                ),
            ]

    def _transform(self, dataset: DataFrame) -> DataFrame:
        out_col = self.getOutputCol()
        input_col = self._validate(self.getInputCol(), dataset)
        path_col = dataset[self.getPathCol()]

        result = dataset.withColumn(
            out_col,
            udf(self.transform_udf, Document.get_schema())(
                input_col,
                path_col,
            ),
        )
        if not self.getKeepInputData():
            result = result.drop(self.getInputCol())
        return result



import logging
import traceback
from types import MappingProxyType
from typing import Any, List

import fitz
import pandas as pd
from pyspark import keyword_only
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.sql.functions import udf

from scaledp.params import (
    HasInputCol,
    HasOutputCol,
    Param,
    Params,
    TypeConverters,
)
from scaledp.schemas.PdfDocument import PdfDocument


class HasGroupByCol(Params):
    """
    Mixin for param groupByCol: column name to group by.
    """

    groupByCol: "Param[str]" = Param(
        Params._dummy(),
        "groupByCol",
        "column name to group by.",
        typeConverter=TypeConverters.toString,
    )

    def __init__(self) -> None:
        super(HasGroupByCol, self).__init__()

    def getGroupByCol(self) -> str:
        """
        Gets the value of groupByCol or its default value.
        """
        return self.getOrDefault(self.groupByCol)

    def setGroupByCol(self, value):
        """
        Sets the value of :py:attr:`groupByCol`.
        """
        return self._set(groupByCol=value)


class PdfAssembler(
    Transformer,
    HasInputCol,
    HasOutputCol,
    HasGroupByCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
):
    """
    Assembles single-page PDFs into a single PDF document.

    Takes a column containing single-page PDF documents, groups them by origin,
    and creates a single PDF using PyMuPDF (fitz).
    """

    DEFAULT_PARAMS = MappingProxyType(
        {
            "inputCol": "pdf",
            "outputCol": "assembled_pdf",
            "groupByCol": "path",
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(PdfAssembler, self).__init__()
        self._setDefault(**self.DEFAULT_PARAMS)
        self._set(**kwargs)

    def convert_to_pdf(self, pdfs: List[PdfDocument]) -> PdfDocument:
        """
        Convert a list of single-page PDF documents into a single PDF.

        Args:
            pdfs: List of PdfDocument objects representing single pages

        Returns:
            PdfDocument: A single PDF document containing all pages
        """
        try:
            if not pdfs or len(pdfs) == 0:
                return PdfDocument(path="", data=bytes(), exception="No PDFs to assemble")

            # Filter out invalid PDFs and sort by page if available
            valid_pdfs = []
            for pdf_page in pdfs:
                if (pdf_page.data is not None
                    and len(pdf_page.data) > 0
                    and pdf_page.exception == ""):
                    valid_pdfs.append(pdf_page)

            if not valid_pdfs:
                return PdfDocument(
                    path=pdfs[0].path if pdfs else "",
                    data=bytes(),
                    exception="No valid PDFs to assemble"
                )

            # Create new PDF document
            pdf = fitz.open()

            for pdf_page in valid_pdfs:
                try:
                    # Open the single page PDF
                    page_doc = fitz.open("pdf", pdf_page.data)
                    # Insert all pages from this document (should be just one)
                    pdf.insert_pdf(page_doc)
                    page_doc.close()
                except Exception as e:
                    logging.warning(f"Failed to insert page from {pdf_page.path}: {e}")
                    continue

            # Write the assembled PDF to bytes
            pdf_bytes = pdf.write()
            pdf.close()

            return PdfDocument(
                path=valid_pdfs[0].path,
                data=pdf_bytes,
                exception=""
            )

        except Exception:
            exception = traceback.format_exc()
            exception = f"PdfAssembler: {exception}"
            logging.warning(exception)
            return PdfDocument(
                path=pdfs[0].path if pdfs and len(pdfs) > 0 else "",
                data=bytes(),
                exception=exception
            )

    def _transform(self, dataset):
        """
        Transform the dataset by grouping single-page PDFs and assembling them.
        """
        output_col = self.getOutputCol()
        input_col = self.getInputCol()
        group_by_col = self.getGroupByCol()

        if input_col not in dataset.columns:
            raise ValueError(f"Input column '{input_col}' is not present in the DataFrame.")

        if group_by_col not in dataset.columns:
            raise ValueError(f"Group by column '{group_by_col}' is not present in the DataFrame.")

        # Check if we're working with pandas DataFrame (PandasPipeline)
        if isinstance(dataset, pd.DataFrame):
            # Pandas DataFrame approach
            # Sort by path and page number if available
            sort_columns = []
            if "path" in dataset.columns:
                sort_columns.append("path")
            if "page_number" in dataset.columns:
                sort_columns.append("page_number")

            if sort_columns:
                dataset = dataset.sort_values(sort_columns).reset_index(drop=True)

            # Group by the specified column and collect PDFs for each group
            grouped = dataset.groupby(group_by_col)[input_col].apply(list).reset_index()
            grouped.columns = [group_by_col, "pdfs"]

            # Apply the conversion function to each group
            assembled_pdfs = []
            for _, row in grouped.iterrows():
                assembled_pdf = self.convert_to_pdf(row["pdfs"])
                assembled_pdfs.append(assembled_pdf)

            # Create result DataFrame
            result = grouped[[group_by_col]].copy()
            result[output_col] = assembled_pdfs

            return result

        else:
            # Spark DataFrame approach (original implementation)
            # Sort by path and page number if available, then group by the specified column
            sorted_dataset = dataset
            if "path" in dataset.columns:
                sorted_dataset = dataset.orderBy("path")
            if "page_number" in dataset.columns:
                sorted_dataset = sorted_dataset.orderBy("page_number")

            # Group by the specified column and collect all PDFs for each group
            from pyspark.sql.functions import collect_list
            grouped_dataset = sorted_dataset.groupBy(dataset[group_by_col]).agg(
                collect_list(dataset[input_col]).alias("pdfs")
            )

            # Apply the UDF to assemble PDFs for each group
            result_dataset = grouped_dataset.withColumn(
                output_col,
                udf(self.convert_to_pdf, PdfDocument.get_schema())("pdfs")
            )

            return result_dataset



"""https://github.com/opencv-ai/paddle-ocr/blob/main/detector.py"""

import numpy as np

from .db_postprocess import DBPostProcess
from .imaug import create_operators, transform
from .predict_base import PredictBase


class DBNetTextDetector(PredictBase):
    def __init__(self, det_model_dir, use_gpu):
        self.det_algorithm = "DB"
        pre_process_list = [
            {
                "DetResizeForTest": {
                    "image_shape": [1280, 1280],
                    # "limit_side_len": 960,
                    # "resize_long": 960,
                    "limit_type": "max",
                    "keep_ratio": True,
                },
            },
            {
                "NormalizeImage": {
                    "std": [0.229, 0.224, 0.225],
                    "mean": [0.485, 0.456, 0.406],
                    "scale": "1./255.",
                    "order": "hwc",
                },
            },
            {"ToCHWImage": None},
            {"KeepKeys": {"keep_keys": ["image", "shape"]}},
        ]
        postprocess_params = {}
        postprocess_params["name"] = "DBPostProcess"
        postprocess_params["thresh"] = 0.5  # args.det_db_thresh
        postprocess_params["box_thresh"] = 0.3  # args.det_db_box_thresh
        postprocess_params["max_candidates"] = 1000
        postprocess_params["unclip_ratio"] = 2.5#1.7  # args.det_db_unclip_ratio
        postprocess_params["use_dilation"] = False  # args.use_dilation
        postprocess_params["score_mode"] = "fast"  # args.det_db_score_mode
        postprocess_params["box_type"] = "quad"  # args.det_box_type
        self.det_box_type = postprocess_params["box_type"]

        self.preprocess_op = create_operators(pre_process_list)
        self.postprocess_op = DBPostProcess(**postprocess_params)

        self.det_onnx_session = self.get_onnx_session(det_model_dir, use_gpu)
        self.det_input_name = self.get_input_name(self.det_onnx_session)
        self.det_output_name = self.get_output_name(self.det_onnx_session)

    def order_points_clockwise(self, pts):
        rect = np.zeros((4, 2), dtype="float32")
        s = pts.sum(axis=1)
        rect[0] = pts[np.argmin(s)]
        rect[2] = pts[np.argmax(s)]
        tmp = np.delete(pts, (np.argmin(s), np.argmax(s)), axis=0)
        diff = np.diff(np.array(tmp), axis=1)
        rect[1] = tmp[np.argmin(diff)]
        rect[3] = tmp[np.argmax(diff)]
        return rect

    def clip_det_res(self, points, img_height, img_width):
        for pno in range(points.shape[0]):
            points[pno, 0] = int(min(max(points[pno, 0], 0), img_width - 1))
            points[pno, 1] = int(min(max(points[pno, 1], 0), img_height - 1))
        return points

    def filter_tag_det_res(self, dt_boxes, image_shape):
        img_height, img_width = image_shape[0:2]
        dt_boxes_new = []
        for box in dt_boxes:
            if type(box) is list:
                box = np.array(box)
            box = self.order_points_clockwise(box)
            box = self.clip_det_res(box, img_height, img_width)
            rect_width = int(np.linalg.norm(box[0] - box[1]))
            rect_height = int(np.linalg.norm(box[0] - box[3]))
            if rect_width <= 3 or rect_height <= 3:
                continue
            dt_boxes_new.append(box)
        dt_boxes = np.array(dt_boxes_new)
        return dt_boxes

    def filter_tag_det_res_only_clip(self, dt_boxes, image_shape):
        img_height, img_width = image_shape[0:2]
        dt_boxes_new = []
        for box in dt_boxes:
            if type(box) is list:
                box = np.array(box)
            box = self.clip_det_res(box, img_height, img_width)
            dt_boxes_new.append(box)
        dt_boxes = np.array(dt_boxes_new)
        return dt_boxes

    def __call__(self, img):
        ori_im = img.copy()
        data = {"image": img}

        data = transform(data, self.preprocess_op)
        img, shape_list = data
        if img is None:
            return None, 0
        img = np.expand_dims(img, axis=0)
        shape_list = np.expand_dims(shape_list, axis=0)
        img = img.copy()

        input_feed = self.get_input_feed(self.det_input_name, img)
        outputs = self.det_onnx_session.run(self.det_output_name, input_feed=input_feed)

        preds = {}
        preds["maps"] = outputs[0]

        post_result = self.postprocess_op(preds, shape_list)
        dt_boxes = post_result[0]["points"]

        if self.det_box_type == "poly":
            dt_boxes = self.filter_tag_det_res_only_clip(dt_boxes, ori_im.shape)
        else:
            dt_boxes = self.filter_tag_det_res(dt_boxes, ori_im.shape)

        return dt_boxes



from dataclasses import dataclass
from typing import Any, Dict

import numpy as np

from scaledp.utils.dataclass import map_dataclass_to_struct, register_type


@dataclass(order=True)
class Box:
    """Box object for represent bounding box data in Spark Dataframe."""

    text: str
    score: float
    x: int
    y: int
    width: int
    height: int
    angle: float = 0.0

    def to_string(self) -> "Box":
        self.text = str(self.text)
        return self

    def json(self) -> Dict[str, Any]:
        return {"text": self.text}

    @staticmethod
    def get_schema():
        return map_dataclass_to_struct(Box)

    def scale(self, factor: float, padding: int = 0) -> "Box":
        return Box(
            text=self.text,
            score=self.score,
            x=int(self.x * factor) - padding,
            y=int(self.y * factor) - padding,
            width=int(self.width * factor) + padding,
            height=int(self.height * factor) + padding,
            angle=self.angle,
        )

    def shape(self, padding: int = 0) -> list[tuple[int, int]]:
        return [
            (self.x - padding, self.y - padding),
            (self.x + self.width + padding, self.y + self.height + padding),
        ]

    def bbox(self, padding: int = 0) -> list[int]:
        return [
            self.x - padding,
            self.y - padding,
            self.x + self.width + padding,
            self.y + self.height + padding,
        ]

    @staticmethod
    def from_bbox(box: list[int], angle: float = 0, label: str = "", score: float = 0):
        return Box(
            text=label,
            score=float(score),
            x=int(box[0]),
            y=int(box[1]),
            width=int(box[2] - box[0]),
            height=int(box[3] - box[1]),
            angle=angle,
        )

    @classmethod
    def from_polygon(
        cls,
        polygon_points: list[tuple[float, float]],
        text: str = "",
        score: float = 1.0,
        padding: int = 0,
    ) -> "Box":
        """
        Creates a Box instance from a list of polygon points (typically 4 for a rectangle).
        Uses OpenCV's minAreaRect to find the rotated bounding box properties.

        Args:
            polygon_points (list[tuple[float, float]]): A list of (x, y) coordinates
                                                        representing the vertices of the polygon.
                                                        Expected to be 4 points for a rectangle.
            text (str): Optional text to assign to the box. Defaults to "".
            score (float): Optional score to assign to the box. Defaults to 1.0.

        Returns:
            Box: A new Box instance representing the minimum enclosing rotated rectangle.

        Raises:
            ValueError: If the number of points is not 4.
        """
        import cv2

        if len(polygon_points) != 4:
            # You might allow more points for convex hull, but for a strict 'Box'
            # (which is a rectangle), 4 points are expected.
            raise ValueError(
                "from_polygon expects exactly 4 points for a rectangular box.",
            )

        # Convert list of tuples to a NumPy array for OpenCV
        points_np = np.array(polygon_points, dtype=np.float32)

        # Get the minimum area rotated rectangle
        (center_x, center_y), (raw_width, raw_height), angle_opencv = cv2.minAreaRect(
            points_np,
        )

        # --- Normalize OpenCV's angle and dimensions to our Box convention ---
        # Our convention: width is the horizontal dimension at 0 degrees, angle 0-360 positive
        # CCW.

        box_width, box_height = raw_width, raw_height
        box_angle = angle_opencv

        # If width is smaller than height, swap dimensions and adjust angle by 90 degrees.
        # This makes 'width' conceptually the longer side or the side oriented towards 0/180
        # degrees.
        if box_width < box_height:
            box_width, box_height = box_height, box_width
            box_angle -= 90.0

        # Normalize angle to 0-360 range, ensuring positive counter-clockwise
        # This handles negative angles from OpenCV and converts to our convention
        box_angle = (box_angle % 360 + 360) % 360

        if box_angle > 270:
            box_angle -= 360

        # --- Derive x, y (top-left of the unrotated box) ---
        # The center is center_x, center_y
        # The top-left of the unrotated box is center - (width/2, height/2)
        # So x = center_x - width/2, y = center_y - height/2

        x = int(round(center_x - box_width / 2.0))
        y = int(round(center_y - box_height / 2.0))

        # Ensure dimensions are integers and positive
        box_width_int = int(round(box_width))
        box_height_int = int(round(box_height))
        box_width_int = max(1, box_width_int)
        box_height_int = max(1, box_height_int)

        return cls(
            text=text,
            score=score,
            x=x - padding,
            y=y - padding,
            width=box_width_int + 2 * padding,
            height=box_height_int + 2 * padding,
            angle=box_angle,
        )

    def is_rotated(self) -> bool:
        return abs(self.angle) >= 3

    @staticmethod
    def iou(box1: "Box", box2: "Box") -> float:
        """Compute Intersection over Union (IoU) between two boxes."""
        x1 = max(box1.x, box2.x)
        y1 = max(box1.y, box2.y)
        x2 = min(box1.x + box1.width, box2.x + box2.width)
        y2 = min(box1.y + box1.height, box2.y + box2.height)
        inter_area = max(0, x2 - x1) * max(0, y2 - y1)
        if inter_area == 0:
            return 0.0
        box1_area = box1.width * box1.height
        box2_area = box2.width * box2.height
        union_area = box1_area + box2_area - inter_area
        return inter_area / union_area

    @staticmethod
    def merge(box1: "Box", box2: "Box") -> "Box":
        """Merge two boxes into one by taking the minimal bounding rectangle."""
        x1 = min(box1.x, box2.x)
        y1 = min(box1.y, box2.y)
        x2 = max(box1.x + box1.width, box2.x + box2.width)
        y2 = max(box1.y + box1.height, box2.y + box2.height)
        return Box(
            text=box1.text or box2.text,
            score=max(box1.score, box2.score),
            x=x1,
            y=y1,
            width=x2 - x1,
            height=y2 - y1,
            angle=0.0,
        )

    @staticmethod
    def is_on_same_line(box1: "Box", box2: "Box", angle_thresh: float = 10.0, line_thresh: float = 0.5) -> bool:
        """
        Check if two boxes are on the same text line.
        - angle_thresh: maximum allowed angle difference (degrees)
        - line_thresh: maximum allowed normalized center difference (as a fraction of height for horizontal text)
        """
        # Check angle similarity
        if abs(box1.angle - box2.angle) > angle_thresh:
            return False
        # For horizontal text (angle near 0)
        if abs(box1.angle) < angle_thresh:
            # Check if vertical centers are close
            y1 = box1.y + box1.height / 2
            y2 = box2.y + box2.height / 2
            avg_height = (box1.height + box2.height) / 2
            return abs(y1 - y2) < avg_height * line_thresh
        else:
            # For rotated text, project centers onto the perpendicular direction
            import math

            theta = math.radians(box1.angle)
            # Direction perpendicular to text line
            dx = -math.sin(theta)
            dy = math.cos(theta)
            c1x = box1.x + box1.width / 2
            c1y = box1.y + box1.height / 2
            c2x = box2.x + box2.width / 2
            c2y = box2.y + box2.height / 2
            # Project difference onto perpendicular direction
            perp_dist = abs((c2x - c1x) * dx + (c2y - c1y) * dy)
            avg_dim = (box1.height + box2.height) / 2
            return perp_dist < avg_dim * line_thresh

    @staticmethod
    def merge_overlapping_boxes(boxes: list["Box"], iou_threshold: float = 0.3, angle_thresh: float = 10.0, line_thresh: float = 0.5) -> list["Box"]:
        """
        Merge all overlapping boxes in a list using a greedy algorithm, but only if they are on the same line and have similar angle.
        """
        merged = []
        used = [False] * len(boxes)
        for i, box in enumerate(boxes):
            if used[i]:
                continue
            curr = box
            for j in range(i + 1, len(boxes)):
                if used[j]:
                    continue
                if (Box.iou(curr, boxes[j]) > iou_threshold and
                    Box.is_on_same_line(curr, boxes[j], angle_thresh=angle_thresh, line_thresh=line_thresh)):
                    curr = Box.merge(curr, boxes[j])
                    used[j] = True
            merged.append(curr)
            used[i] = True
        return merged


register_type(Box, Box.get_schema)



import json
from types import MappingProxyType
from typing import Any

from pyspark import keyword_only

from scaledp.params import HasLLM, HasPrompt, HasSchema
from scaledp.schemas.ExtractorOutput import ExtractorOutput
from scaledp.utils.pydantic_shema_utils import json_schema_to_model

from .BaseVisualExtractor import BaseVisualExtractor


class GeminiVisualExtractor(BaseVisualExtractor, HasLLM, HasSchema, HasPrompt):

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "data",
            "keepInputData": True,
            "model": "gemini-2.5-flash",
            "apiBase": "",
            "apiKey": "",
            "numPartitions": 1,
            "pageCol": "page",
            "pathCol": "path",
            "prompt": """Please extract data from the scanned image as json.""",
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(GeminiVisualExtractor, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.pipeline = None

    def call_extractor(self, images, params):
        import base64

        import google.generativeai as genai

        schema = json.loads(params["schema"])
        schema = json_schema_to_model(schema, schema.get("$defs", {}))
        generation_config = {
            "temperature": 1,
            "top_p": 0.95,
            "top_k": 40,
            "max_output_tokens": 8192,
            "response_mime_type": "application/json",
            "response_schema": schema,
        }

        genai.configure(api_key=params["apiKey"])

        model = genai.GenerativeModel(
            model_name=params["model"],
            generation_config=generation_config,
        )

        schema = json.loads(params["schema"])
        schema = json_schema_to_model(schema, schema.get("$defs", {}))

        results = []

        for image in images:
            image_decoded = base64.b64encode(image.data).decode("utf-8")
            prompt = params["prompt"].format(schema=schema)
            response = model.generate_content(
                [{"mime_type": "image/png", "data": image_decoded}, prompt],
                stream=False,
            )
            results.append(
                ExtractorOutput(
                    path=image.path,
                    data=response.text,
                    type="GeminiVisualExtractor",
                    exception="",
                ),
            )
        return results



import json
import os
from types import MappingProxyType
from typing import Any

from pyspark import keyword_only

from ...params import HasLLM, HasSchema
from ...schemas.ExtractorOutput import ExtractorOutput
from ...utils.pydantic_shema_utils import json_schema_to_model
from .BaseExtractor import BaseExtractor


class DSPyExtractor(BaseExtractor, HasLLM, HasSchema):

    defaultParams = MappingProxyType(
        {
            "inputCol": "text",
            "outputCol": "data",
            "keepInputData": True,
            "model": "llama3-8b-8192",
            "apiBase": None,
            "apiKey": None,
            "numPartitions": 1,
            "pageCol": "page",
            "pathCol": "path",
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(DSPyExtractor, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.pipeline = None

    def call_extractor(self, documents, params):
        import dspy

        lm = dspy.LM(
            params["model"],
            api_base=params["apiBase"] or os.environ["OPENAI_BASE_URL"],
            api_key=params["apiKey"] or os.environ["OPENAI_API_KEY"],
        )
        dspy.configure(lm=lm)

        schema = json.loads(params["schema"])
        schema = json_schema_to_model(schema, schema.get("$defs", {}))

        class ExtractData(dspy.Signature):
            """improve a recognized text and Extract structured information from the receipt."""

            text: str = dspy.InputField(
                desc="""OCR recognized text of the receipt from the Ukrainian store""",
            )
            data: schema = dspy.OutputField(
                desc="Structured data from the receipt with fixes "
                "and improvements of OCR recognition.",
            )

        module = dspy.ChainOfThought(ExtractData)

        results = []
        for document in documents:
            data = module(text=document.text).data
            results.append(
                ExtractorOutput(
                    path=document.path,
                    data=data.json_data(),
                    type="DSPyExtractor",
                    exception="",
                ),
            )
        return results



import gc
import logging
from types import MappingProxyType
from typing import Any

import numpy as np
from pyspark import keyword_only

from scaledp.enums import Device
from scaledp.models.detectors.BaseDetector import BaseDetector
from scaledp.params import HasBatchSize, HasDevice, HasModel, HasWhiteList
from scaledp.schemas.Box import Box
from scaledp.schemas.DetectorOutput import DetectorOutput


class LayoutDetector(BaseDetector, HasDevice, HasBatchSize, HasWhiteList, HasModel):
    _model = None

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "layout_boxes",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.5,
            "device": Device.CPU,
            "batchSize": 2,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "propagateError": False,
            "onlyRotated": False,
            "model": "PP-DocLayout_plus-L",
            "whiteList": [],
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(LayoutDetector, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.get_model({k.name: v for k, v in self.extractParamMap().items()})

    @classmethod
    def get_model(cls, params):
        logging.info("Loading PaddleOCR LayoutDetection model...")
        if cls._model:
            return cls._model

        try:
            from paddleocr import LayoutDetection
        except ImportError as e:
            raise ImportError(
                "PaddleOCR is not installed. Please install it with: pip install paddleocr",
            ) from e

        # Get model name from params or use default
        model_name = params.get("model", "PP-DocLayout_plus-L")

        # Initialize LayoutDetection model
        device = "gpu" if int(params["device"]) == Device.CUDA.value else "cpu"
        cls._model = LayoutDetection(
            model_name=model_name,
            enable_hpi=False,
            device=device,
        )

        logging.info(
            f"PaddleOCR LayoutDetection model '{model_name}' loaded successfully",
        )
        return cls._model

    @classmethod
    def call_detector(cls, images, params):
        logging.info("Running LayoutDetector")

        detector = cls.get_model(params)
        layout_types = params.get("whiteList", [])

        logging.info("Process images for layout detection")
        results_final = []

        for image, image_path in images:
            boxes = []

            # Convert PIL to NumPy (RGB)
            image_np = np.array(image)

            try:
                # Run layout analysis using LayoutDetection
                result = detector.predict(input=image_np)

                if result and len(result) > 0:
                    # LayoutDetection returns a list of layout regions
                    result = result[0]
                    if isinstance(result, dict) and "boxes" in result:
                        for layout_item in result["boxes"]:
                            bbox = layout_item["coordinate"]  # Bounding box coordinates
                            layout_type = layout_item["label"]  # Layout type
                            confidence = layout_item.get("score", 1.0)

                            # Filter by layout type if specified
                            if layout_types and layout_type not in layout_types:
                                continue

                            # Filter by confidence threshold
                            if confidence < params["scoreThreshold"]:
                                continue

                            # Convert bbox to Box format
                            # LayoutDetection returns bbox as [x1, y1, x2, y2]
                            if len(bbox) == 4:
                                x = bbox[0]
                                y = bbox[1]
                                width = bbox[2] - bbox[0]
                                height = bbox[3] - bbox[1]

                                # Create Box with layout type as text
                                box = Box(
                                    text=layout_type,
                                    score=confidence,
                                    x=int(x),
                                    y=int(y),
                                    width=int(width),
                                    height=int(height),
                                )

                                # Add polygon points if needed for rotated boxes
                                if len(bbox) == 4:
                                    box.polygon = bbox

                                boxes.append(box)

            except Exception as e:
                logging.warning(f"Error in layout detection for {image_path}: {e!s}")
                if params.get("propagateError", False):
                    raise e

            if params.get("onlyRotated", False):
                boxes = [box for box in boxes if box.is_rotated()]

            results_final.append(
                DetectorOutput(path=image_path, type="layout", bboxes=boxes),
            )

        gc.collect()

        return results_final



import io
import logging
import traceback
from types import MappingProxyType
from typing import Any

import fitz
from pyspark import keyword_only
from pyspark.ml import Transformer
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.sql.functions import udf

from scaledp.params import HasOutputCol, HasInputCols, HasColumnValidator, \
    HasResolution
from scaledp.schemas.PdfDocument import PdfDocument


class PdfAddTextLayer(
    Transformer,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasOutputCol,
    HasInputCols,
    HasColumnValidator,
    HasResolution,
):
    """Add text layer to PDF document using text from Document schema."""

    DEFAULT_PARAMS = MappingProxyType(
        {
            "inputCols": ["pdf", "text"],
            "outputCol": "pdf_with_text",
            "resolution": 300,  # DPI for coordinate transformation
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(PdfAddTextLayer, self).__init__()
        self._setDefault(**self.DEFAULT_PARAMS)
        self._set(**kwargs)

    def transform_udf(self, pdf_doc, text_doc):
        """Transform PDF and text documents to create PDF with text layer."""
        try:
            # Check for exceptions in input documents
            if pdf_doc.exception != "":
                return PdfDocument(
                    path=pdf_doc.path,
                    data=bytes(),
                    width=pdf_doc.width,
                    height=pdf_doc.height,
                    exception=pdf_doc.exception,
                )

            if text_doc.exception != "":
                return PdfDocument(
                    path=pdf_doc.path,
                    data=bytes(),
                    width=pdf_doc.width,
                    height=pdf_doc.height,
                    exception=text_doc.exception,
                )

            if not pdf_doc.data:
                return PdfDocument(
                    path=pdf_doc.path,
                    data=bytes(),
                    width=pdf_doc.width,
                    height=pdf_doc.height,
                    exception="PDF document has no data",
                )

            # Open the PDF document
            pdf_document = fitz.open(stream=pdf_doc.data, filetype="pdf")

            if len(pdf_document) == 0:
                pdf_document.close()
                return PdfDocument(
                    path=pdf_doc.path,
                    data=bytes(),
                    width=pdf_doc.width,
                    height=pdf_doc.height,
                    exception="PDF document has no pages",
                )

            # Get the first page (assuming single page PDF as per requirement)
            page = pdf_document[0]

            # Get page dimensions and calculate coordinate transformation
            page_width = page.mediabox_size[0]  # PDF points
            page_height = page.mediabox_size[1]  # PDF points

            # Calculate scale factor from image coordinates to PDF coordinates
            pdf_dpi = 72.0     # PDF native DPI
            scale_factor = pdf_dpi / self.getResolution()

            # Add text layer using bounding boxes from Document
            if text_doc.bboxes:
                for bbox in text_doc.bboxes:
                    # Convert image coordinates to PDF coordinates
                    # Image coordinates: origin at top-left, y increases downward
                    # PDF coordinates: origin at bottom-left, y increases upward

                    pdf_x = bbox.x * scale_factor
                    pdf_y = (bbox.y - 0.2 * bbox.height) * scale_factor  # Flip Y axis
                    pdf_width = bbox.width * scale_factor
                    pdf_height = bbox.height * scale_factor

                    # Insert text at the specified position using the correct PyMuPDF method
                    page.insert_text(
                        point=fitz.Point(pdf_x, pdf_y + pdf_height),  # Bottom-left of text
                        text=bbox.text,
                        fontsize=max(8, pdf_height * 0.8),  # Scale font size
                        # with bbox height
                        color=(0, 0, 0),  # Black text
                        overlay=False,
                    )


            # Save the modified PDF to bytes
            output_buffer = io.BytesIO()
            pdf_document.save(output_buffer)
            pdf_bytes = output_buffer.getvalue()
            pdf_document.close()

            return PdfDocument(
                path=pdf_doc.path,
                data=pdf_bytes,
                width=pdf_doc.width,
                height=pdf_doc.height,
                exception="",
            )

        except Exception:
            exception = traceback.format_exc()
            exception = f"PdfAddTextLayer: {exception}"
            logging.warning(exception)
            return PdfDocument(
                path=pdf_doc.path if pdf_doc else "",
                data=bytes(),
                width=pdf_doc.width if pdf_doc else None,
                height=pdf_doc.height if pdf_doc else None,
                exception=exception,
            )

    def _transform(self, dataset):
        """Transform the dataset by adding text layer to PDF documents."""
        output_col = self.getOutputCol()
        input_cols = self.getInputCols()

        # Validate that we have exactly 2 input columns
        if len(input_cols) != 2:
            raise ValueError(
                f"PdfAddTextLayer requires exactly 2 input columns (PDF and text), got {len(input_cols)}",
            )

        pdf_col, text_col = input_cols

        # Validate input columns exist
        if pdf_col not in dataset.columns:
            raise ValueError(
                f"PDF input column '{pdf_col}' is not present in the DataFrame.",
            )

        if text_col not in dataset.columns:
            raise ValueError(
                f"Text input column '{text_col}' is not present in the DataFrame.",
            )

        pdf_column = dataset[pdf_col]
        text_column = dataset[text_col]

        return dataset.withColumn(
            output_col,
            udf(self.transform_udf, PdfDocument.get_schema())(pdf_column, text_column),
        )



from types import MappingProxyType
from typing import Any

import numpy
from pyspark import keyword_only

from scaledp.enums import Device
from scaledp.models.detectors.BaseDetector import BaseDetector
from scaledp.params import HasBatchSize, HasDevice, Param, Params, TypeConverters
from scaledp.schemas.Box import Box
from scaledp.schemas.DetectorOutput import DetectorOutput


class CraftTextDetector(BaseDetector, HasDevice, HasBatchSize):
    _craft_net = None
    _refine_net = None

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "boxes",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.7,
            "textThreshold": 0.4,
            "linkThreshold": 0.4,
            "sizeThreshold": -1,
            "width": 1280,
            "withRefiner": False,
            "device": Device.CPU,
            "batchSize": 2,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "propagateError": False,
            "onlyRotated": False,
        },
    )

    textThreshold = Param(
        Params._dummy(),
        "textThreshold",
        "Threshold for text region score",
        typeConverter=TypeConverters.toFloat,
    )

    linkThreshold = Param(
        Params._dummy(),
        "linkThreshold",
        "Threshold for link affinity score",
        typeConverter=TypeConverters.toFloat,
    )

    sizeThreshold = Param(
        Params._dummy(),
        "sizeThreshold",
        "Threshold for height of detected regions",
        typeConverter=TypeConverters.toInt,
    )

    width = Param(
        Params._dummy(),
        "width",
        "Width for image resizing",
        typeConverter=TypeConverters.toInt,
    )

    withRefiner = Param(
        Params._dummy(),
        "withRefiner",
        "Enable refiner network postprocessing",
        typeConverter=TypeConverters.toBoolean,
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(CraftTextDetector, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.get_model({k.name: v for k, v in self.extractParamMap().items()})

    @classmethod
    def get_model(cls, params):
        if cls._craft_net and cls._refine_net:
            return cls._craft_net, cls._refine_net

        from craft_text_detector import load_craftnet_model, load_refinenet_model

        device = "cuda" if int(params["device"]) == Device.CUDA.value else "cpu"
        use_cuda = device == "cuda"

        craft_net = load_craftnet_model(
            cuda=use_cuda,
        )

        refine_net = None
        if params.get("withRefiner"):
            refine_net = load_refinenet_model(
                cuda=use_cuda,
            )

        cls._craft_net = craft_net
        cls._refine_net = refine_net
        return craft_net, refine_net

    @classmethod
    def call_detector(cls, images, params):
        import cv2
        from craft_text_detector import craft_utils, image_utils, torch_utils

        craft_net, refine_net = cls.get_model(params)
        use_cuda = int(params["device"]) == Device.CUDA.value
        results = []

        for image, image_path in images:
            try:
                # Convert PIL to OpenCV format

                img_cv = image_utils.read_image(numpy.array(image)[:, :, ::-1])

                # Resize
                img_resized, target_ratio, size_heatmap = (
                    image_utils.resize_aspect_ratio(
                        img_cv,
                        params["width"],
                        interpolation=cv2.INTER_LINEAR,
                    )
                )
                ratio_h = ratio_w = 1 / target_ratio

                # Preprocess
                x = image_utils.normalizeMeanVariance(img_resized)
                x = torch_utils.from_numpy(x).permute(2, 0, 1)
                x = torch_utils.Variable(x.unsqueeze(0))
                if use_cuda:
                    x = x.cuda()

                # Forward pass
                with torch_utils.no_grad():
                    y, feature = craft_net(x)

                score_text = y[0, :, :, 0].cpu().data.numpy()
                score_link = y[0, :, :, 1].cpu().data.numpy()

                # Refine if enabled
                if refine_net is not None:
                    with torch_utils.no_grad():
                        y_refiner = refine_net(y, feature)
                    score_link = y_refiner[0, :, :, 0].cpu().data.numpy()

                # Post-process
                boxes, _ = craft_utils.getDetBoxes(
                    score_text,
                    score_link,
                    params["scoreThreshold"],
                    params["linkThreshold"],
                    params["textThreshold"],
                    poly=False,
                )

                # Adjust coordinates
                boxes = craft_utils.adjustResultCoordinates(boxes, ratio_w, ratio_h)

                # Convert to Box objects
                box_objects = [Box.from_polygon(box) for box in boxes]

                results.append(
                    DetectorOutput(
                        path=image_path,
                        type="craft",
                        bboxes=box_objects,
                        exception="",
                    ),
                )

            except Exception as e:
                raise e
                results.append(
                    DetectorOutput(
                        path=image_path,
                        type="craft",
                        bboxes=[],
                        exception=f"CraftTextDetector error: {e!s}",
                    ),
                )

        return results

    @classmethod
    def call_detector1(cls, images, params):
        import cv2
        from crafter import craft_utils, image_utils

        craft_net, refine_net = cls.get_model(params)
        results = []

        for image, image_path in images:
            try:
                # Convert PIL to OpenCV format

                img_cv = image_utils.read_image(numpy.array(image)[:, :, ::-1])

                # Resize
                img_resized, target_ratio, size_heatmap = (
                    image_utils.resize_aspect_ratio(
                        img_cv,
                        params["width"],
                        interpolation=cv2.INTER_LINEAR,
                    )
                )
                ratio_h = ratio_w = 1 / target_ratio

                # Preprocess
                x = image_utils.normalizeMeanVariance(img_resized)
                x = numpy.transpose(x, (2, 0, 1))  # [h, w, c] to [c, h, w]
                x = numpy.expand_dims(x, 0)  # [c, h, w] to [b, c, h, w]

                # Forward pass
                y, feature = craft_net(x)

                score_text = y[0, 0, :, :]
                score_link = y[0, 1, :, :]

                # Refine if enabled
                if refine_net is not None:
                    y_refiner = refine_net(y, feature)
                    score_link = y_refiner[0, 0, :, :]

                # Post-process
                boxes, _ = craft_utils.getDetBoxes(
                    score_text,
                    score_link,
                    params["scoreThreshold"],
                    params["linkThreshold"],
                    params["textThreshold"],
                    poly=False,
                )

                # Adjust coordinates
                boxes = craft_utils.adjustResultCoordinates(boxes, ratio_w, ratio_h)

                # Convert to Box objects
                box_objects = [Box.from_polygon(box) for box in boxes]

                results.append(
                    DetectorOutput(
                        path=image_path,
                        type="craft",
                        bboxes=box_objects,
                        exception="",
                    ),
                )

            except Exception as e:
                raise e
                results.append(
                    DetectorOutput(
                        path=image_path,
                        type="craft",
                        bboxes=[],
                        exception=f"CraftTextDetector error: {e!s}",
                    ),
                )

        return results

    def setTextThreshold(self, value):
        return self._set(textThreshold=value)

    def setLinkThreshold(self, value):
        return self._set(linkThreshold=value)

    def setSizeThreshold(self, value):
        return self._set(sizeThreshold=value)

    def setWidth(self, value):
        return self._set(width=value)

    def setWithRefiner(self, value):
        return self._set(withRefiner=value)



import gc
import logging
from pathlib import Path
from types import MappingProxyType
from typing import Any

import numpy as np
from huggingface_hub import hf_hub_download
from pyspark import keyword_only

from scaledp.enums import Device
from scaledp.models.detectors.BaseDetector import BaseDetector
from scaledp.models.detectors.paddle_onnx.predict_det import DBNetTextDetector
from scaledp.params import HasBatchSize, HasDevice
from scaledp.schemas.Box import Box
from scaledp.schemas.DetectorOutput import DetectorOutput


class DBNetOnnxDetector(BaseDetector, HasDevice, HasBatchSize):
    _model = None

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "boxes",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.2,
            "device": Device.CPU,
            "batchSize": 2,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "propagateError": False,
            "onlyRotated": False
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(DBNetOnnxDetector, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.get_model({k.name: v for k, v in self.extractParamMap().items()})

    @classmethod
    def get_model(cls, params):

        logging.info("Loading model...")
        if cls._model:
            return cls._model

        model = params["model"]
        if not Path(model).is_file():
            model = hf_hub_download(repo_id=model, filename="model.onnx")

        logging.info("Model downloaded")

        detector = DBNetTextDetector(model, use_gpu=params["model"] == Device.CUDA)

        cls._model = detector
        return cls._model

    @classmethod
    def call_detector(cls, images, params):
        logging.info("Running DBNetOnnxDetector")
        import cv2

        detector = cls.get_model(params)

        logging.info("Process images")
        results_final = []
        for image, image_path in images:
            boxes = []

            # Convert PIL to NumPy (RGB)
            image_np = np.array(image)

            # Convert RGB to BGR for OpenCV
            image_rgb = cv2.cvtColor(image_np, cv2.COLOR_RGB2BGR)

            result = detector(image_rgb)

            for points in result:
                boxes.append(Box.from_polygon(points, padding=0))

            if params["onlyRotated"]:
                boxes = [box for box in boxes if box.is_rotated()]

            # Merge overlapping boxes before returning, only if on the same line and similar angle
            boxes = Box.merge_overlapping_boxes(
                boxes,
                iou_threshold=0.02,  # as before
                angle_thresh=10.0,  # only merge if angle difference < 10 degrees
                line_thresh=0.3     # only merge if centers are close (half
                # height)
            )

            results_final.append(
                DetectorOutput(path=image_path, type="DBNetOnnx", bboxes=boxes),
            )

        gc.collect()

        return results_final



"""
This code is refered from:
https://github.com/WenmuZhou/DBNet.pytorch/blob/master/post_processing/seg_detector_representer.py
"""

import cv2
import numpy as np
import pyclipper
from shapely.geometry import Polygon


class DBPostProcess(object):
    """
    The post process for Differentiable Binarization (DB).
    """

    def __init__(
        self,
        thresh=0.3,
        box_thresh=0.7,
        max_candidates=1000,
        unclip_ratio=2.0,
        use_dilation=False,
        score_mode="fast",
        box_type="quad",
        **kwargs
    ):
        self.thresh = thresh
        self.box_thresh = box_thresh
        self.max_candidates = max_candidates
        self.unclip_ratio = unclip_ratio
        self.min_size = 3
        self.score_mode = score_mode
        self.box_type = box_type
        assert score_mode in [
            "slow",
            "fast",
        ], "Score mode must be in [slow, fast] but got: {}".format(score_mode)

        self.dilation_kernel = None if not use_dilation else np.array([[1, 1], [1, 1]])

    def polygons_from_bitmap(self, pred, _bitmap, dest_width, dest_height):
        """
        _bitmap: single map with shape (1, H, W),
            whose values are binarized as {0, 1}
        """

        bitmap = _bitmap
        height, width = bitmap.shape

        boxes = []
        scores = []

        contours, _ = cv2.findContours(
            (bitmap * 255).astype(np.uint8), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE
        )

        for contour in contours[: self.max_candidates]:
            epsilon = 0.002 * cv2.arcLength(contour, True)
            approx = cv2.approxPolyDP(contour, epsilon, True)
            points = approx.reshape((-1, 2))
            if points.shape[0] < 4:
                continue

            score = self.box_score_fast(pred, points.reshape(-1, 2))
            if self.box_thresh > score:
                continue

            if points.shape[0] > 2:
                box = self.unclip(points, self.unclip_ratio)
                if len(box) > 1:
                    continue
            else:
                continue
            box = box.reshape(-1, 2)

            _, sside = self.get_mini_boxes(box.reshape((-1, 1, 2)))
            if sside < self.min_size + 2:
                continue

            box = np.array(box)
            box[:, 0] = np.clip(np.round(box[:, 0] / width * dest_width), 0, dest_width)
            box[:, 1] = np.clip(
                np.round(box[:, 1] / height * dest_height), 0, dest_height
            )
            boxes.append(box.tolist())
            scores.append(score)
        return boxes, scores

    def boxes_from_bitmap(
        self, pred, _bitmap, dest_width, dest_height, ratio_h, ratio_w
    ):
        """
        _bitmap: single map with shape (1, H, W),
                whose values are binarized as {0, 1}
        """

        bitmap = _bitmap
        height, width = bitmap.shape

        outs = cv2.findContours(
            (bitmap * 255).astype(np.uint8), cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE
        )
        if len(outs) == 3:
            img, contours, _ = outs[0], outs[1], outs[2]
        elif len(outs) == 2:
            contours, _ = outs[0], outs[1]

        num_contours = min(len(contours), self.max_candidates)

        boxes = []
        scores = []
        for index in range(num_contours):
            contour = contours[index]
            points, sside = self.get_mini_boxes(contour)
            if sside < self.min_size:
                continue
            points = np.array(points)
            if self.score_mode == "fast":
                score = self.box_score_fast(pred, points.reshape(-1, 2))
            else:
                score = self.box_score_slow(pred, contour)
            if self.box_thresh > score:
                continue

            box = self.unclip(points, self.unclip_ratio).reshape(-1, 1, 2)
            box, sside = self.get_mini_boxes(box)
            if sside < self.min_size + 2:
                continue
            box = np.array(box)

            # box[:, 0] = np.clip(
            #     np.round(box[:, 0] / width * dest_width), 0, dest_width)
            # box[:, 1] = np.clip(
            #     np.round(box[:, 1] / height * dest_height), 0, dest_height)

            box[:, 0] = np.clip(np.round(box[:, 0] / ratio_w), 0, dest_width)
            box[:, 1] = np.clip(np.round(box[:, 1] / ratio_h), 0, dest_height)
            boxes.append(box.astype("int32"))
            scores.append(score)
        return np.array(boxes, dtype="int32"), scores

    def unclip(self, box, unclip_ratio):
        poly = Polygon(box)
        distance = poly.area * unclip_ratio / poly.length
        offset = pyclipper.PyclipperOffset()
        offset.AddPath(box, pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
        expanded = np.array(offset.Execute(distance))
        return expanded

    def get_mini_boxes(self, contour):
        bounding_box = cv2.minAreaRect(contour)
        points = sorted(list(cv2.boxPoints(bounding_box)), key=lambda x: x[0])

        index_1, index_2, index_3, index_4 = 0, 1, 2, 3
        if points[1][1] > points[0][1]:
            index_1 = 0
            index_4 = 1
        else:
            index_1 = 1
            index_4 = 0
        if points[3][1] > points[2][1]:
            index_2 = 2
            index_3 = 3
        else:
            index_2 = 3
            index_3 = 2

        box = [
            points[index_1],
            points[index_2],
            points[index_3],
            points[index_4],
        ]
        return box, min(bounding_box[1])

    def box_score_fast(self, bitmap, _box):
        """
        box_score_fast: use bbox mean score as the mean score
        """
        h, w = bitmap.shape[:2]
        box = _box.copy()
        xmin = np.clip(np.floor(box[:, 0].min()).astype("int32"), 0, w - 1)
        xmax = np.clip(np.ceil(box[:, 0].max()).astype("int32"), 0, w - 1)
        ymin = np.clip(np.floor(box[:, 1].min()).astype("int32"), 0, h - 1)
        ymax = np.clip(np.ceil(box[:, 1].max()).astype("int32"), 0, h - 1)

        mask = np.zeros((ymax - ymin + 1, xmax - xmin + 1), dtype=np.uint8)
        box[:, 0] = box[:, 0] - xmin
        box[:, 1] = box[:, 1] - ymin
        cv2.fillPoly(mask, box.reshape(1, -1, 2).astype("int32"), 1)
        return cv2.mean(bitmap[ymin : ymax + 1, xmin : xmax + 1], mask)[0]

    def box_score_slow(self, bitmap, contour):
        """
        box_score_slow: use polyon mean score as the mean score
        """
        h, w = bitmap.shape[:2]
        contour = contour.copy()
        contour = np.reshape(contour, (-1, 2))

        xmin = np.clip(np.min(contour[:, 0]), 0, w - 1)
        xmax = np.clip(np.max(contour[:, 0]), 0, w - 1)
        ymin = np.clip(np.min(contour[:, 1]), 0, h - 1)
        ymax = np.clip(np.max(contour[:, 1]), 0, h - 1)

        mask = np.zeros((ymax - ymin + 1, xmax - xmin + 1), dtype=np.uint8)

        contour[:, 0] = contour[:, 0] - xmin
        contour[:, 1] = contour[:, 1] - ymin

        cv2.fillPoly(mask, contour.reshape(1, -1, 2).astype("int32"), 1)
        return cv2.mean(bitmap[ymin : ymax + 1, xmin : xmax + 1], mask)[0]

    def __call__(self, outs_dict, shape_list):
        pred = outs_dict["maps"]
        pred = pred[:, 0, :, :]
        segmentation = pred > self.thresh

        boxes_batch = []
        for batch_index in range(pred.shape[0]):
            src_h, src_w, ratio_h, ratio_w = shape_list[batch_index]
            if self.dilation_kernel is not None:
                mask = cv2.dilate(
                    np.array(segmentation[batch_index]).astype(np.uint8),
                    self.dilation_kernel,
                )
            else:
                mask = segmentation[batch_index]
            if self.box_type == "poly":
                boxes, scores = self.polygons_from_bitmap(
                    pred[batch_index], mask, src_w, src_h
                )
            elif self.box_type == "quad":
                boxes, scores = self.boxes_from_bitmap(
                    pred[batch_index], mask, src_w, src_h, ratio_h, ratio_w
                )
            else:
                raise ValueError("box_type can only be one of ['quad', 'poly']")

            boxes_batch.append({"points": boxes})
        return boxes_batch


class DistillationDBPostProcess(object):
    def __init__(
        self,
        model_name=["student"],
        key=None,
        thresh=0.3,
        box_thresh=0.6,
        max_candidates=1000,
        unclip_ratio=1.5,
        use_dilation=False,
        score_mode="fast",
        box_type="quad",
        **kwargs
    ):
        self.model_name = model_name
        self.key = key
        self.post_process = DBPostProcess(
            thresh=thresh,
            box_thresh=box_thresh,
            max_candidates=max_candidates,
            unclip_ratio=unclip_ratio,
            use_dilation=use_dilation,
            score_mode=score_mode,
            box_type=box_type,
        )

    def __call__(self, predicts, shape_list):
        results = {}
        for k in self.model_name:
            results[k] = self.post_process(predicts[k], shape_list=shape_list)
        return results



import onnxruntime
from crafter.resources import res


class Craftnet:
    def __init__(self, onnx_path=None):
        onnx_path = "/home/mykola/PycharmProjects/scaledp/tests/model.quant.onnx"

        session_options = onnxruntime.SessionOptions()
        if onnx_path is None:
            onnx_path = res("craftnet.onnx")
        self._onnx_session = onnxruntime.InferenceSession(
            onnx_path,
            sess_options=session_options,
        )

    def __call__(self, image):
        return self._onnx_session.run(None, {"image": image})


class Refinenet:
    def __init__(self, onnx_path=None):
        if onnx_path is None:
            onnx_path = res("refinenet.onnx")
        self._onnx_session = onnxruntime.InferenceSession(onnx_path)

    def __call__(self, y, feature):
        return self._onnx_session.run(None, {"y": y, "feature": feature})[0]



import gc
import logging
from pathlib import Path
from types import MappingProxyType
from typing import Any

import numpy as np
from huggingface_hub import hf_hub_download
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters

from scaledp.enums import Device
from scaledp.models.detectors.BaseDetector import BaseDetector
from scaledp.models.detectors.yolo.yolo import YOLO
from scaledp.params import HasBatchSize, HasDevice
from scaledp.schemas.Box import Box
from scaledp.schemas.DetectorOutput import DetectorOutput


class YoloOnnxDetector(BaseDetector, HasDevice, HasBatchSize):
    _model = None

    task = Param(
        Params._dummy(),
        "task",
        "Yolo task type.",
        typeConverter=TypeConverters.toString,
    )

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "boxes",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.2,
            "device": Device.CPU,
            "batchSize": 2,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "propagateError": False,
            "task": "detect",
            "onlyRotated": False,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(YoloOnnxDetector, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.get_model({k.name: v for k, v in self.extractParamMap().items()})

    @classmethod
    def get_model(cls, params):

        logging.info("Loading model...")
        if cls._model:
            return cls._model

        model = params["model"]
        if not Path(model).is_file():
            model = hf_hub_download(repo_id=model, filename="model.onnx")

        logging.info("Model downloaded")

        detector = YOLO(model, params["scoreThreshold"])

        cls._model = detector
        return cls._model

    @classmethod
    def call_detector(cls, images, params):
        logging.info("Running YoloOnnxDetector")
        detector = cls.get_model(params)

        logging.info("Process images")
        results_final = []
        for image, image_path in images:
            boxes = []
            # Convert PIL to NumPy (RGB)
            image_np = np.array(image)
            raw_boxes, scores, class_ids = detector.detect_objects(image_np)

            for box in raw_boxes:
                boxes.append(Box.from_bbox(box))
            results_final.append(
                DetectorOutput(path=image_path, type="yolo", bboxes=boxes),
            )

        gc.collect()

        return results_final



import logging
from types import MappingProxyType
from typing import Any

import numpy as np
import torch
from doctr.models import detection_predictor
from pyspark import keyword_only

from scaledp.enums import Device
from scaledp.models.detectors.BaseDetector import BaseDetector
from scaledp.params import HasBatchSize, HasDevice
from scaledp.schemas.Box import Box
from scaledp.schemas.DetectorOutput import DetectorOutput


class FastTextDetector(BaseDetector, HasDevice, HasBatchSize):
    _model = None
    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "boxes",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.7,
            "batchSize": 2,
            "device": Device.CPU,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "propagateError": False,
            "onlyRotated": False,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(FastTextDetector, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.get_model({k.name: v for k, v in self.extractParamMap().items()})

    @classmethod
    def get_model(cls, params):
        if cls._model:
            return cls._model
        device = "cuda" if int(params["device"]) == Device.CUDA.value else "cpu"

        # Initialize the fast text detection model from doctr
        cls._model = detection_predictor(
            arch="fast_tiny",  # Use DBNet with ResNet50 backbone
            pretrained=True,  # Use pretrained weights
            assume_straight_pages=True,
        ).to(device)

        return cls._model

    @classmethod
    def call_detector(cls, images, params):
        model = cls.get_model(params)
        device = "cuda" if int(params["device"]) == Device.CUDA.value else "cpu"
        results = []

        for img, image_path in images:
            try:
                # Convert PIL image to numpy array if needed
                image = img
                if not isinstance(image, np.ndarray):
                    image = np.array(image)

                # Normalize image
                if image.max() > 1:
                    image = image / 255.0

                # Get predictions
                out = model([[image]])

                # Extract boxes from predictions
                # doctr returns relative coordinates (0-1)
                predictions = out[0]

                # Convert doctr boxes to Box objects
                # Scale relative coordinates to absolute coordinates
                height, width = image.shape[:2]
                box_objects = []
                for pred in predictions:
                    # Get coordinates
                    x_min, y_min = pred[0]  # Top-left
                    x_max, y_max = pred[2]  # Bottom-right

                    # Convert to absolute coordinates
                    abs_coords = [
                        int(x_min * width),  # x_min
                        int(y_min * height),  # y_min
                        int(x_max * width),  # x_max
                        int(y_max * height),  # y_max
                    ]

                    box_objects.append(Box.from_bbox(abs_coords))
                results.append(
                    DetectorOutput(
                        path=image_path,
                        type="fast",
                        bboxes=box_objects,
                        exception="",
                    ),
                )

            except Exception as e:
                logging.exception(e)
                results.append(
                    DetectorOutput(
                        path=image_path,
                        type="fast",
                        bboxes=[],
                        exception=f"FastTextDetector error: {e!s}",
                    ),
                )

        if device == "cuda":
            torch.cuda.empty_cache()

        return results



import gc
import logging
from pathlib import Path
from types import MappingProxyType
from typing import Any

import numpy as np
from huggingface_hub import hf_hub_download
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters

from scaledp.enums import Device
from scaledp.models.detectors.BaseDetector import BaseDetector
from scaledp.params import HasBatchSize, HasDevice
from scaledp.schemas.Box import Box
from scaledp.schemas.DetectorOutput import DetectorOutput


class YoloOnnxTextDetector(BaseDetector, HasDevice, HasBatchSize):
    _model = None

    task = Param(
        Params._dummy(),
        "task",
        "Yolo task type.",
        typeConverter=TypeConverters.toString,
    )

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "boxes",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.2,
            "device": Device.CPU,
            "batchSize": 2,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "propagateError": False,
            "task": "detect",
            "onlyRotated": False,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(YoloOnnxTextDetector, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.get_model({k.name: v for k, v in self.extractParamMap().items()})

    @classmethod
    def get_model(cls, params):

        logging.info("Loading model...")
        if cls._model:
            return cls._model

        import onnxruntime as ort

        model = params["model"]
        if not Path(model).is_file():
            model = hf_hub_download(repo_id=model, filename="model.onnx")

        logging.info("Model downloaded")

        detector = ort.InferenceSession(model, providers=["CPUExecutionProvider"])
        logging.info("Ort session created")

        cls._model = detector
        return cls._model

    @classmethod
    def obb_iou(cls, rect1, rect2):
        import cv2

        int_pts = cv2.rotatedRectangleIntersection(rect1, rect2)[1]
        if int_pts is None:
            return 0.0
        int_area = cv2.contourArea(int_pts)
        area1 = rect1[1][0] * rect1[1][1]
        area2 = rect2[1][0] * rect2[1][1]
        union = area1 + area2 - int_area
        return int_area / union if union > 0 else 0.0

    @classmethod
    def obb_nms(cls, boxes, iou_thresh=0.3):
        # boxes: [N, 6] -> cx, cy, w, h, conf, angle
        boxes = sorted(boxes, key=lambda x: x[4], reverse=True)  # sort by confidence
        keep = []

        while boxes:
            current = boxes.pop(0)
            keep.append(current)

            current_rect = (
                (current[0], current[1]),
                (current[2], current[3]),
                np.degrees(current[5]),
            )

            remaining = []
            for box in boxes:
                other_rect = (
                    (box[0], box[1]),
                    (box[2], box[3]),
                    np.degrees(box[5]),
                )
                iou = cls.obb_iou(current_rect, other_rect)
                if iou < iou_thresh:
                    remaining.append(box)
            boxes = remaining

        return np.array(keep)

    @classmethod
    def call_detector(cls, images, params):
        logging.info("Running YoloOnnxDetector")
        import cv2

        detector = cls.get_model(params)

        logging.info("Process images")
        results_final = []
        for image, image_path in images:
            boxes = []

            # Convert PIL to NumPy (RGB)
            image_np = np.array(image)

            # Convert RGB to BGR for OpenCV
            image_rgb = cv2.cvtColor(image_np, cv2.COLOR_RGB2BGR)

            target_h = 960
            target_w = 960

            scale = min(target_w / image_rgb.shape[1], target_h / image_rgb.shape[0])
            resize_h = int(image_rgb.shape[0] * scale)
            resize_w = int(image_rgb.shape[1] * scale)
            img = cv2.resize(image_rgb, (resize_w, resize_h))

            if resize_h < target_h:
                pad_bottom = target_h - resize_h
                image_resized = cv2.copyMakeBorder(
                    img,
                    0,
                    pad_bottom,
                    0,
                    0,
                    cv2.BORDER_CONSTANT,
                    value=(255, 255, 255),
                )
            if resize_w < target_w:
                pad_right = target_w - resize_w
                image_resized = cv2.copyMakeBorder(
                    img,
                    0,
                    0,
                    0,
                    pad_right,
                    cv2.BORDER_CONSTANT,
                    value=(255, 255, 255),
                )
            ratio_h = ratio_w = scale

            # Preprocess
            input_tensor = image_resized.astype(np.float32) / 255.0  # Normalize
            input_tensor = np.transpose(input_tensor, (2, 0, 1))  # HWC -> CHW
            input_tensor = np.expand_dims(input_tensor, axis=0)  # Add batch dim

            input_name = detector.get_inputs()[0].name

            logging.info("Run ort inference")

            # Inference
            outputs = detector.run(None, {input_name: input_tensor})

            logging.info("Complete ort inference")

            raw = outputs[0].squeeze(0)  # shape: (6, 8400)

            # Step 2: Transpose to (8400, 6)
            output = raw.T  # shape: (8400, 6)

            # Step 3: Confidence mask
            mask = output[:, 4] > params["scoreThreshold"]
            filtered = output[mask]

            nms_result = cls.obb_nms(filtered, iou_thresh=0.2)

            if params["task"] == "obb":

                for obb in nms_result:
                    cx, cy, bw, bh, conf, angle = obb

                    # Scale to original image if needed (skip if already absolute)
                    angle_scaled = float(
                        np.degrees(angle),
                    )  # Convert radians to degrees
                    box = ((float(cx), float(cy)), (float(bw), float(bh)), angle_scaled)
                    points = cv2.boxPoints(box)
                    points[:, 0] *= ratio_w  # x coordinates
                    points[:, 1] *= ratio_h  # y coordinates
                    points = np.intp(points)
                    boxes.append(Box.from_polygon(points))
            else:
                for box in nms_result:
                    boxes.append(Box.from_bbox(box.xyxy[0]))
            results_final.append(
                DetectorOutput(path=image_path, type="yolo", bboxes=boxes),
            )

        gc.collect()

        return results_final



import gc
from pathlib import Path
from types import MappingProxyType
from typing import Any

from huggingface_hub import hf_hub_download
from huggingface_hub.errors import EntryNotFoundError
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters

from scaledp.models.detectors.BaseDetector import BaseDetector
from scaledp.params import HasBatchSize, HasDevice
from scaledp.schemas.Box import Box
from scaledp.schemas.DetectorOutput import DetectorOutput

from ...enums import Device


class YoloDetector(BaseDetector, HasDevice, HasBatchSize):
    _model = None

    task = Param(
        Params._dummy(),
        "task",
        "Yolo task type.",
        typeConverter=TypeConverters.toString,
    )

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "boxes",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.5,
            "device": Device.CPU,
            "batchSize": 2,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "propagateError": False,
            "task": "detect",
            "onlyRotated": False,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(YoloDetector, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.get_model({k.name: v for k, v in self.extractParamMap().items()})

    @classmethod
    def get_model(cls, params):
        if cls._model:
            return cls._model
        from ultralytics import YOLO

        model = params["model"]
        if not Path(model).is_file():
            try:
                model = hf_hub_download(repo_id=model, filename="best.pt")
            except EntryNotFoundError:
                model = hf_hub_download(repo_id=model, filename="model.onnx")

        detector = YOLO(model, task=params["task"])
        device = "cpu" if int(params["device"]) == Device.CPU.value else "cuda"
        if "onnx" not in model:
            cls._model = detector.to(device)
        else:
            cls._model = detector
        return cls._model

    @classmethod
    def call_detector(cls, images, params):
        import torch

        detector = cls.get_model(params)

        results = detector(
            [image[0] for image in images],
            conf=params["scoreThreshold"],
            save_conf=True,
        )

        results_final = []
        for res, (_image, image_path) in zip(results, images):
            boxes = []

            if params["task"] == "obb":

                for obb in res.obb:
                    xyxyxyxy = obb.xyxyxyxy.to("cpu").numpy().astype(int)
                    boxes.append(Box.from_polygon(xyxyxyxy.tolist()[0]))
            else:
                for box in res.boxes:
                    boxes.append(Box.from_bbox(box.xyxy[0]))
            results_final.append(
                DetectorOutput(path=image_path, type="yolo", bboxes=boxes),
            )

        gc.collect()
        if int(params["device"]) == Device.CUDA.value:
            torch.cuda.empty_cache()

        return results_final



"""
https://github.com/PaddlePaddle/PaddleOCR/blob/main/ppocr/data/imaug/__init__.py
"""

from .operators import *


def transform(data, ops=None):
    """Transform"""
    if ops is None:
        ops = []
    for op in ops:
        data = op(data)
        if data is None:
            return None
    return data


def create_operators(op_param_list, global_config=None):
    """
    create operators based on the config

    Args:
        params(list): a dict list, used to create some operators
    """
    assert isinstance(op_param_list, list), "operator config should be a list"
    ops = []
    for operator in op_param_list:
        assert isinstance(operator, dict) and len(operator) == 1, "yaml format error"
        op_name = list(operator)[0]
        param = {} if operator[op_name] is None else operator[op_name]
        if global_config is not None:
            param.update(global_config)
        op = eval(op_name)(**param)
        ops.append(op)
    return ops



import numpy as np


def nms(boxes, scores, iou_threshold):
    # Sort by score
    sorted_indices = np.argsort(scores)[::-1]

    keep_boxes = []
    while sorted_indices.size > 0:
        # Pick the last box
        box_id = sorted_indices[0]
        keep_boxes.append(box_id)

        # Compute IoU of the picked box with the rest
        ious = compute_iou(boxes[box_id, :], boxes[sorted_indices[1:], :])

        # Remove boxes with IoU over the threshold
        keep_indices = np.where(ious < iou_threshold)[0]

        # print(keep_indices.shape, sorted_indices.shape)
        sorted_indices = sorted_indices[keep_indices + 1]

    return keep_boxes


def multiclass_nms(boxes, scores, class_ids, iou_threshold):

    unique_class_ids = np.unique(class_ids)

    keep_boxes = []
    for class_id in unique_class_ids:
        class_indices = np.where(class_ids == class_id)[0]
        class_boxes = boxes[class_indices, :]
        class_scores = scores[class_indices]

        class_keep_boxes = nms(class_boxes, class_scores, iou_threshold)
        keep_boxes.extend(class_indices[class_keep_boxes])

    return keep_boxes


def compute_iou(box, boxes):
    """Compute xmin, ymin, xmax, ymax for both boxes."""
    xmin = np.maximum(box[0], boxes[:, 0])
    ymin = np.maximum(box[1], boxes[:, 1])
    xmax = np.minimum(box[2], boxes[:, 2])
    ymax = np.minimum(box[3], boxes[:, 3])

    # Compute intersection area
    intersection_area = np.maximum(0, xmax - xmin) * np.maximum(0, ymax - ymin)

    # Compute union area
    box_area = (box[2] - box[0]) * (box[3] - box[1])
    boxes_area = (boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1])
    union_area = box_area + boxes_area - intersection_area

    # Compute IoU
    iou = intersection_area / union_area

    return iou


def xywh2xyxy(x):
    """Convert bounding box (x, y, w, h) to bounding box (x1, y1, x2, y2)."""
    y = np.copy(x)
    y[..., 0] = x[..., 0] - x[..., 2] / 2
    y[..., 1] = x[..., 1] - x[..., 3] / 2
    y[..., 2] = x[..., 0] + x[..., 2] / 2
    y[..., 3] = x[..., 1] + x[..., 3] / 2
    return y



from dataclasses import dataclass

from scaledp.utils.dataclass import BinaryT, map_dataclass_to_struct, register_type


@dataclass(order=True)
class PdfDocument:
    path: str
    data: BinaryT
    width: int = None
    height: int = None
    exception: str = ""

    @staticmethod
    def get_schema():
        return map_dataclass_to_struct(PdfDocument)


register_type(PdfDocument, PdfDocument.get_schema)



from dataclasses import dataclass

from scaledp.schemas.Entity import Entity
from scaledp.utils.dataclass import map_dataclass_to_struct, register_type


@dataclass(order=True)
class NerOutput:
    path: str
    entities: list[Entity]
    exception: str
    json: str = ""

    @staticmethod
    def get_schema():
        return map_dataclass_to_struct(NerOutput)


register_type(NerOutput, NerOutput.get_schema)



from pathlib import Path
from typing import Any

from huggingface_hub import hf_hub_download
from pyspark.ml.param import Param, Params, TypeConverters
import numpy as np
import cv2
import onnxruntime as ort

class HasDetectLineOrientation(Params):
    """
    Mixin for param detectLineOrientation: whether to detect line orientation, and logic for orientation detection.
    """
    detectLineOrientation = Param(
        Params._dummy(),
        "detectLineOrientation",
        "Whether to detect line orientation.",
        typeConverter=TypeConverters.toBoolean,
    )

    oriModel = Param(
        Params._dummy(),
        "oriModel",
        "Text line Orientation Model.",
        typeConverter=TypeConverters.toString,
    )


    def getOriModel(self) -> str:
        """
        Gets the value of model or its default value.
        """
        return self.getOrDefault(self.model)

    def setOriModel(self, value: str) -> Any:
        """
        Sets the value of :py:attr:`model`.
        """
        return self._set(model=value)

    _orientation_session = None
    _orientation_input_name = None
    _orientation_label_list = ["0_degree", "180_degree"]

    def setDetectLineOrientation(self, value: bool):
        return self._set(detectLineOrientation=value)

    def getDetectLineOrientation(self) -> bool:
        return self.getOrDefault(self.detectLineOrientation)

    @classmethod
    def _load_orientation_model(cls, params):
        if cls._orientation_session is None:
            model_path = params.get("oriModel")
            if not Path(model_path).is_file():
                model = hf_hub_download(repo_id=model_path, filename="model.onnx")
            cls._orientation_session = ort.InferenceSession(model)
            cls._orientation_input_name = cls._orientation_session.get_inputs()[0].name

    @staticmethod
    def _preprocess_for_orientation(pil_img):
        img = np.array(pil_img)
        img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
        img = cv2.resize(img, (160, 80))  # width, height
        img = img.astype(np.float32) / 255.0
        mean = np.array([0.485, 0.456, 0.406])
        std = np.array([0.229, 0.224, 0.225])
        img = (img - mean) / std
        img = np.transpose(img, (2, 0, 1))  # HWC -> CHW
        img = np.expand_dims(img, 0)  # batch
        return img.astype(np.float32)

    @classmethod
    def detect_orientation(cls, pil_img, params):
        """Detects orientation (0 or 180 degrees) of a PIL image."""
        cls._load_orientation_model(params)
        inp = cls._preprocess_for_orientation(pil_img)
        outputs = cls._orientation_session.run(None, {cls._orientation_input_name: inp})
        pred_idx = np.argmax(outputs[0], axis=1)[0]
        pred_label = cls._orientation_label_list[pred_idx]
        return pred_label

    @classmethod
    def auto_orient_image(cls, pil_img, params):
        """Rotates the image to 0 degrees if needed."""
        orientation = cls.detect_orientation(pil_img, params)
        if orientation == "180_degree":
            return pil_img.rotate(180, expand=True), orientation
        return pil_img, orientation






import onnxruntime


class PredictBase(object):
    def __init__(self):
        pass

    def get_onnx_session(self, model_dir, use_gpu):
        # 使用gpu
        if use_gpu:
            providers = [
                "CUDAExecutionProvider"
            ]  # [('CUDAExecutionProvider',{"cudnn_conv_algo_search": "DEFAULT"})]#,'CPUExecutionProvider']
        else:
            providers = ["CPUExecutionProvider"]

        session_options = onnxruntime.SessionOptions()
        session_options.intra_op_num_threads = 1  # Set number of CPU cores
        onnx_session = onnxruntime.InferenceSession(
            model_dir, session_options, providers=providers
        )

        # print("providers:", onnxruntime.get_device())
        return onnx_session

    def get_output_name(self, onnx_session):
        """
        output_name = onnx_session.get_outputs()[0].name
        :param onnx_session:
        :return:
        """
        output_name = []
        for node in onnx_session.get_outputs():
            output_name.append(node.name)
        return output_name

    def get_input_name(self, onnx_session):
        """
        input_name = onnx_session.get_inputs()[0].name
        :param onnx_session:
        :return:
        """
        input_name = []
        for node in onnx_session.get_inputs():
            input_name.append(node.name)
        return input_name

    def get_input_feed(self, input_name, image_numpy):
        """
        input_feed={self.input_name: image_numpy}
        :param input_name:
        :param image_numpy:
        :return:
        """
        input_feed = {}
        for name in input_name:
            input_feed[name] = image_numpy
        return input_feed



import logging
from types import MappingProxyType
from typing import Any

import cv2
import numpy as np
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters

from scaledp.params import CODE_TO_LANGUAGE, LANGUAGE_TO_TESSERACT_CODE
from scaledp.models.detectors.HasDetectLineOrientation import HasDetectLineOrientation
from scaledp.schemas.Box import Box
from scaledp.schemas.Document import Document

from ...enums import OEM, PSM, TessLib
from .BaseRecognizer import BaseRecognizer


class TesseractRecognizer(BaseRecognizer, HasDetectLineOrientation):
    """
    Run Tesseract text recognition on images.
    """

    oem = Param(
        Params._dummy(),
        "oem",
        "OCR engine mode. Defaults to :attr:`OEM.DEFAULT`.",
        typeConverter=TypeConverters.toInt,
    )

    tessDataPath = Param(
        Params._dummy(),
        "tessDataPath",
        "Path to tesseract data folder.",
        typeConverter=TypeConverters.toString,
    )

    tessLib = Param(
        Params._dummy(),
        "tessLib",
        "The desired Tesseract library to use. Defaults to :attr:`TESSEROCR`",
        typeConverter=TypeConverters.toInt,
    )

    onlyRotated = Param(
        Params._dummy(),
        "onlyRotated",
        "Return only rotated boxes.",
        typeConverter=TypeConverters.toBoolean,
    )

    defaultParams = MappingProxyType(
        {
            "inputCols": ["image", "boxes"],
            "outputCol": "text",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.5,
            "oem": OEM.DEFAULT,
            "lang": ["eng"],
            "lineTolerance": 0,
            "keepFormatting": False,
            "tessDataPath": "/usr/share/tesseract-ocr/5/tessdata/",
            "tessLib": TessLib.PYTESSERACT,
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "detectLineOrientation": True,
            "onlyRotated": True,
            "oriModel": "StabRise/line_orientation_detection_v0.1",
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(TesseractRecognizer, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    @staticmethod
    def getLangTess(params):
        return "+".join(
            [
                LANGUAGE_TO_TESSERACT_CODE[CODE_TO_LANGUAGE[lang]]
                for lang in params["lang"]
            ],
        )

    @classmethod
    def _prepare_box_for_ocr(cls, image_np, box, params):


        scaled_box = box.scale(params["scaleFactor"], padding=5)
        # if scaled_box.angle == 90:
        #     scaled_box.angle = -90

        center_tuple = (
            scaled_box.x + scaled_box.width / 2,
            scaled_box.y + scaled_box.height / 2,
        )
        size_tuple = (scaled_box.width, scaled_box.height)

        rect = (center_tuple, size_tuple, scaled_box.angle)
        src_pts = cv2.boxPoints(rect).astype("float32")
        dst_pts = np.array(
            [
                [0, int(scaled_box.height) - 1],
                [0, 0],
                [int(scaled_box.width) - 1, 0],
                [int(scaled_box.width) - 1, int(scaled_box.height) - 1],
            ],
            dtype="float32",
        )

        try:
            m_transform = cv2.getPerspectiveTransform(src_pts, dst_pts)
            return cv2.warpPerspective(
                image_np,
                m_transform,
                (int(scaled_box.width), int(scaled_box.height)),
                flags=cv2.INTER_CUBIC,
                borderMode=cv2.BORDER_REPLICATE,
            )
        except cv2.error as e:
            logging.error(f"Error processing box {box}: {e}")
            return None

    @classmethod
    def _convert_to_pil(cls, cropped_np):
        from PIL import Image

        if cropped_np is None or cropped_np.size == 0:
            return None
        if cropped_np.ndim == 3:
            if cropped_np.shape[2] == 4:
                cropped_np = cv2.cvtColor(cropped_np, cv2.COLOR_RGBA2RGB)
            elif cropped_np.shape[2] not in (3,):
                cropped_np = cv2.cvtColor(cropped_np, cv2.COLOR_BGR2GRAY)
        elif cropped_np.ndim != 2:
            return None
        return Image.fromarray(cropped_np)

    @classmethod
    def _process_image_with_tesseract(cls, image, image_path, detected_boxes, params):
        from tesserocr import PSM, PyTessBaseAPI



        boxes, texts = [], []
        image_np = np.array(image)
        lang = cls.getLangTess(params)

        with (PyTessBaseAPI(
            path=params["tessDataPath"],
            psm=PSM.SINGLE_WORD,
            oem=params["oem"],
            lang=lang,
        ) as api):
            api.SetVariable("debug_file", "ocr.log")
            for box in detected_boxes.bboxes:
                # Ensure box is Box instance
                if isinstance(box, dict):
                    box = Box(**box)
                elif not isinstance(box, Box):
                    box = Box(**box.asDict())
                cropped_np = cls._prepare_box_for_ocr(image_np, box, params)
                # Auto-orient the image before OCR

                pil_image = cls._convert_to_pil(cropped_np)
                if params["detectLineOrientation"]:
                    pil_image, orientation = cls.auto_orient_image(pil_image,
                                                                   params)
                if pil_image is None:
                    continue

                if params["onlyRotated"] and not box.is_rotated() and \
                    orientation != "180_degree":
                    continue

                api.SetImage(pil_image)
                api.Recognize(0)
                b = box
                if isinstance(box, dict):
                    b = Box(**b)
                b.text = api.GetUTF8Text()
                b.conf = api.MeanTextConf()
                if b.score > params["scoreThreshold"]:
                    boxes.append(b)
                    texts.append(b.text)

        if params["keepFormatting"]:
            text = TesseractRecognizer.box_to_formatted_text(
                boxes,
                params["lineTolerance"],
            )
        else:
            text = " ".join(texts)

        return Document(
            path=image_path,
            text=text,
            bboxes=boxes,
            type="text",
            exception="",
        )

    @classmethod
    def call_pytesseract(cls, images, detected_boxes, params):  # pragma: no cover
        results = []
        for (image, image_path), boxes in zip(images, detected_boxes):
            doc = cls._process_image_with_tesseract(image, image_path, boxes, params)
            results.append(doc)
        return results

    @classmethod
    def call_tesserocr(cls, images, detected_boxes, params):  # pragma: no cover
        from tesserocr import PyTessBaseAPI

        results = []
        lang = cls.getLangTess(params)
        with PyTessBaseAPI() as api:
            api.Init(params["tessDataPath"], lang, oem=params["oem"])
            api.SetPageSegMode(PSM.SINGLE_WORD)
            api.SetVariable("debug_file", "ocr.log")

            for (image, image_path), detected_box in zip(images, detected_boxes):
                api.SetImage(image)

                boxes = []
                texts = []

                for b in detected_box.bboxes:
                    box = b
                    if isinstance(box, dict):
                        box = Box(**box)
                    if not isinstance(box, Box):
                        box = Box(**box.asDict())
                    scaled_box = box.scale(params["scaleFactor"], padding=0)
                    api.SetRectangle(
                        scaled_box.x,
                        scaled_box.y,
                        scaled_box.width,
                        scaled_box.height,
                    )
                    box.text = api.GetUTF8Text().replace("\n", "")
                    box.score = api.MeanTextConf() / 100
                    if box.score > params["scoreThreshold"]:
                        boxes.append(box)
                        texts.append(box.text)
                if params["keepFormatting"]:
                    text = TesseractRecognizer.box_to_formatted_text(
                        boxes,
                        params["lineTolerance"],
                    )
                else:
                    text = " ".join(texts)

                results.append(
                    Document(
                        path=image_path,
                        text=text,
                        bboxes=boxes,
                        type="text",
                        exception="",
                    ),
                )
        return results

    @classmethod
    def call_recognizer(cls, images, boxes, params):
        if params["tessLib"] == TessLib.TESSEROCR.value:
            return cls.call_tesserocr(images, boxes, params)
        if params["tessLib"] == TessLib.PYTESSERACT.value:
            return cls.call_pytesseract(images, boxes, params)
        raise ValueError(f"Unknown Tesseract library: {params['tessLib']}")

    def setOem(self, value):
        """
        Sets the value of :py:attr:`oem`.
        """
        return self._set(oem=value)

    def getOem(self):
        """
        Sets the value of :py:attr:`oem`.
        """
        return self.getOrDefault(self.oem)

    def setTessDataPath(self, value):
        """
        Sets the value of :py:attr:`tessDataPath`.
        """
        return self._set(tessDataPath=value)

    def getTessDataPath(self):
        """
        Sets the value of :py:attr:`tessDataPath`.
        """
        return self.getOrDefault(self.tessDataPath)

    def setTessLib(self, value):
        """
        Sets the value of :py:attr:`tessLib`.
        """
        return self._set(tessLib=value)

    def getTessLib(self):
        """
        Gets the value of :py:attr:`tessLib`.
        """
        return self.getOrDefault(self.tessLib)



import io
import logging
from dataclasses import dataclass

import imagesize
from PIL import Image as pImage

from scaledp.utils.dataclass import BinaryT, map_dataclass_to_struct, register_type

from ..enums import ImageType


@dataclass(order=True)
class Image(object):
    """Image object for represent image data in Spark Dataframe."""

    path: str
    resolution: int = 0
    data: BinaryT = bytes()
    imageType: str = ImageType.FILE.value
    exception: str = ""
    height: int = 0
    width: int = 0

    def to_pil(self) -> pImage.Image:
        """Convert image to PIL Image format."""
        if self.imageType in (ImageType.FILE.value, ImageType.WEBP.value):
            return pImage.open(io.BytesIO(self.data))
        raise ValueError("Invalid image type.")

    def to_cv2(self):
        """Convert image to OpenCV format."""
        import cv2
        import numpy as np

        if self.imageType in (ImageType.FILE.value, ImageType.WEBP.value):
            # Convert bytes to numpy array
            nparr = np.frombuffer(self.data, np.uint8)
            # Decode the numpy array as an image
            return cv2.imdecode(nparr, cv2.IMREAD_COLOR)
        raise ValueError("Invalid image type.")

    def to_io_stream(self) -> io.BytesIO:
        return io.BytesIO(self.data)

    def to_webp(self) -> "Image":
        if self.imageType == ImageType.FILE.value:
            image = pImage.open(io.BytesIO(self.data))
            buff = io.BytesIO()
            image.save(buff, "webp")
            self.data = buff.getvalue()
        return self

    @staticmethod
    def from_binary(data, path, imageType, resolution=None, width=None, height=None):
        img = Image(
            path=path,
            data=data,
            imageType=ImageType.FILE.value,
            resolution=resolution,
        )
        if data is None or len(data) == 0:
            raise ValueError("Empty image data.")
        if imageType in (ImageType.FILE.value, ImageType.WEBP.value):
            if height is not None:
                img.height = height
            if width is not None:
                img.width = width
            if width is None and height is None:
                img.width, img.height = imagesize.get(io.BytesIO(img.data))
                if img.width == -1:
                    raise Exception("Unable to read image.")
                logging.info(f"Image size: {img.width}x{img.height}")
        return img

    @staticmethod
    def from_pil(data, path, imageType, resolution):
        buff = io.BytesIO()
        if imageType == ImageType.WEBP.value:
            data.save(buff, "webp")
        else:
            data.save(buff, "png")
        return Image(
            path=path,
            data=buff.getvalue(),
            imageType=ImageType.FILE.value,
            width=data.width,
            height=data.height,
            resolution=resolution,
        )

    @staticmethod
    def get_schema():
        return map_dataclass_to_struct(Image)

    def __str__(self) -> str:
        return (
            f"Image(path={self.path}, resolution={self.resolution}, "
            f"imageType={self.imageType}, exception={self.exception}, "
            f"height={self.height}, width={self.width})"
        )

    def __repr__(self) -> str:
        return self.__str__()


register_type(Image, Image.get_schema)



from enum import Enum, IntEnum


class ImageType(Enum):
    FILE = "file"
    OPENCV = "opencv"
    PIL = "pil"
    WEBP = "webp"


class Device(IntEnum):
    CPU = -1
    CUDA = 0
    CUDA_0 = 0
    CUDA_1 = 1
    CUDA_2 = 2


class PSM(IntEnum):
    AUTO = 3
    AUTO_ONLY = 2
    AUTO_OSD = 1
    CIRCLE_WORD = 9
    COUNT = 14
    OSD_ONLY = 0
    RAW_LINE = 13
    SINGLE_BLOCK = 6
    SINGLE_BLOCK_VERT_TEXT = 5
    SINGLE_CHAR = 10
    SINGLE_COLUMN = 4
    SINGLE_LINE = 7
    SINGLE_WORD = 8
    SPARSE_TEXT = 11
    SPARSE_TEXT_OSD = 12


class OEM(IntEnum):
    DEFAULT = 3
    LSTM_ONLY = 1
    TESSERACT_LSTM_COMBINED = 2
    TESSERACT_ONLY = 0


class TessLib(IntEnum):
    TESSEROCR = 0
    PYTESSERACT = 1



import json
import logging
from types import MappingProxyType
from typing import Any

from pyspark import keyword_only
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)

from ...params import HasLLM, HasPrompt, HasSchema
from ...schemas.ExtractorOutput import ExtractorOutput
from .BaseExtractor import BaseExtractor


class LLMExtractor(BaseExtractor, HasLLM, HasSchema, HasPrompt):

    defaultParams = MappingProxyType(
        {
            "inputCol": "text",
            "outputCol": "data",
            "keepInputData": True,
            "model": "llama3-8b-8192",
            "apiBase": None,
            "apiKey": None,
            "numPartitions": 1,
            "pageCol": "page",
            "pathCol": "path",
            "prompt": """Please extract data from the text as json.""",
            "systemPrompt": "You are data extractor from the scanned images.",
            "delay": 30,
            "maxRetry": 6,
            "propagateError": False,
            "temperature": 1.0,
            "schemaByPrompt": True,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(LLMExtractor, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.pipeline = None

    def call_extractor(self, documents, params):
        from openai import RateLimitError

        client = self.getOIClient()

        @retry(
            retry=retry_if_exception_type(RateLimitError),
            wait=wait_random_exponential(min=1, max=self.getDelay()),
            stop=stop_after_attempt(self.getMaxRetry()),
        )
        def completion_with_backoff(**kwargs: Any):
            logging.info("Calling LLM API")
            return client.beta.chat.completions.parse(**kwargs)

        results = []
        for document in documents:
            data = document.text

            content = [{"type": "text", "text": params["prompt"]}]
            kwargs = {}

            if self.getSchemaByPrompt():
                content.append(
                    {
                        "type": "text",
                        "text": "Schema for the output json: "
                        + self.getSchema()
                        + " Always return valid json. Do not include schema to the output.",
                    },
                )
            else:
                kwargs["response_format"] = self.getPaydanticSchema()

            completion = completion_with_backoff(
                model=params["model"],
                messages=[
                    {
                        "role": "user",
                        "content": [
                            {"type": "text", "text": data},
                        ],
                    },
                    {"role": "user", "content": content},
                ],
                temperature=self.getTemperature(),
                **kwargs,
            )
            result = (
                completion.choices[0]
                .message.content.replace("```json", "")
                .replace("```", "")
            )

            results.append(
                ExtractorOutput(
                    path=document.path,
                    data=json.dumps(json.loads(result), indent=4, ensure_ascii=False),
                    type="LLMExtractor",
                    exception="",
                ),
            )
        return results



import json
import os
from enum import Enum
from typing import Any, List

from pydantic import BaseModel
from pyspark.ml.param import Param, Params, TypeConverters

from scaledp.langs import (
    CODE_TO_LANGUAGE,
    LANGUAGE_TO_CODE,
    LANGUAGE_TO_TESSERACT_CODE,
    TESSERACT_CODE_TO_LANGUAGE,
)
from scaledp.utils.pydantic_shema_utils import json_schema_to_model


class AutoParamsMeta(type(Params), type):
    def __new__(cls, name, bases, dct) -> Any:
        dct_copy = dct.copy()
        for attr_name, attr_value in dct_copy.items():
            if isinstance(attr_value, Param):
                set_method_name = f"set{attr_name[0].upper()}{attr_name[1:]}"
                get_method_name = f"get{attr_name[0].upper()}{attr_name[1:]}"

                def set_method(self, value, attr_name=attr_name):
                    return self._set(**{attr_name: value})

                def get_method(self, attr_name=attr_name):
                    return self.getOrDefault(attr_name)

                dct[set_method_name] = set_method
                dct[get_method_name] = get_method

        return super(AutoParamsMeta, cls).__new__(cls, name, bases, dct)


class HasImageType(Params):

    imageType = Param(
        Params._dummy(),
        "imageType",
        "Image type.",
        typeConverter=TypeConverters.toString,
    )

    def setImageType(self, value):
        """
        Sets the value of :py:attr:`imageType`.
        """
        return self._set(imageType=value)

    def getImageType(self):
        """
        Sets the value of :py:attr:`imageType`.
        """
        return self.getOrDefault(self.imageType)


class HasKeepInputData(Params):

    keepInputData = Param(
        Params._dummy(),
        "keepInputData",
        "Keep input data column in output.",
        typeConverter=TypeConverters.toBoolean,
    )

    def setKeepInputData(self, value):
        """
        Sets the value of :py:attr:`keepInputData`.
        """
        return self._set(keepInputData=value)

    def getKeepInputData(self):
        """
        Sets the value of :py:attr:`keepInputData`.
        """
        return self.getOrDefault(self.keepInputData)


class HasPathCol(Params):
    """
    Mixin for param pathCol: path column name.
    """

    pathCol = Param(
        Params._dummy(),
        "pathCol",
        "Input column name with path of file.",
        typeConverter=TypeConverters.toString,
    )

    def setPathCol(self, value):
        """
        Sets the value of :py:attr:`pathCol`.
        """
        return self._set(pathCol=value)

    def getPathCol(self) -> str:
        """
        Gets the value of pathCol or its default value.
        """
        return self.getOrDefault(self.pathCol)


class HasInputCols(Params):
    """
    Mixin for param inputCols: input column names.
    """

    inputCols: "Param[List[str]]" = Param(
        Params._dummy(),
        "inputCols",
        "input column names.",
        typeConverter=TypeConverters.toListString,
    )

    def __init__(self) -> None:
        super(HasInputCols, self).__init__()

    def getInputCols(self):
        """
        Gets the value of inputCols or its default value.
        """
        return self.getOrDefault(self.inputCols)

    def setInputCols(self, value):
        """
        Sets the value of :py:attr:`inputCol`.
        """
        return self._set(inputCols=value)


class HasInputCol(Params):
    """
    Mixin for param inputCol: input column name.
    """

    inputCol: "Param[str]" = Param(
        Params._dummy(),
        "inputCol",
        "input column name.",
        typeConverter=TypeConverters.toString,
    )

    def __init__(self) -> None:
        super(HasInputCol, self).__init__()

    def getInputCol(self) -> str:
        """
        Gets the value of inputCol or its default value.
        """
        return self.getOrDefault(self.inputCol)

    def setInputCol(self, value):
        """
        Sets the value of :py:attr:`inputCol`.
        """
        return self._set(inputCol=value)


class HasOutputCol(Params):
    """
    Mixin for param outputCol: output column name.
    """

    outputCol: "Param[str]" = Param(
        Params._dummy(),
        "outputCol",
        "output column name.",
        typeConverter=TypeConverters.toString,
    )

    def __init__(self) -> None:
        super(HasOutputCol, self).__init__()
        self._setDefault(outputCol=self.uid + "__output")

    def getOutputCol(self) -> str:
        """
        Gets the value of outputCol or its default value.
        """
        return self.getOrDefault(self.outputCol)

    def setOutputCol(self, value):
        """
        Sets the value of :py:attr:`outputCol`.
        """
        return self._set(outputCol=value)


class HasBypassCol(Params):
    """
    Mixin for param inputCol: input column name.
    """

    bypassCol: "Param[str]" = Param(
        Params._dummy(),
        "bypassCol",
        "Bypass input column name.",
        typeConverter=TypeConverters.toString,
    )

    def __init__(self) -> None:
        super(HasBypassCol, self).__init__()

    def getBypassCol(self) -> str:
        """
        Gets the value of bypassCol or its default value.
        """
        return self.getOrDefault(self.bypassCol)

    def setBypassCol(self, value):
        """
        Sets the value of :py:attr:`bypassCol`.
        """
        return self._set(bypassCol=value)


class HasResolution(Params):
    resolution = Param(
        Params._dummy(),
        "resolution",
        "Resolution of image.",
        typeConverter=TypeConverters.toInt,
    )

    POINTS_PER_INCH = 72

    def setResolution(self, value):
        """
        Sets the value of :py:attr:`resolution`.
        """
        return self._set(resolution=value)

    def getResolution(self):
        """
        Gets the value of :py:attr:`resolution`.
        """
        return self.getOrDefault(self.resolution)


class HasPageCol(Params):
    """
    Mixin for param pageCol: path column name.
    """

    pageCol = Param(
        Params._dummy(),
        "pageCol",
        "Page column name.",
        typeConverter=TypeConverters.toString,
    )

    def setPageCol(self, value):
        """
        Sets the value of :py:attr:`pageCol`.
        """
        return self._set(pageCol=value)

    def getPageCol(self) -> str:
        """
        Gets the value of pageCol or its default value.
        """
        return self.getOrDefault(self.pageCol)


class HasNumPartitions:
    numPartitions = Param(
        Params._dummy(),
        "numPartitions",
        "Number of partitions.",
        typeConverter=TypeConverters.toInt,
    )

    def setNumPartitions(self, value):
        """
        Sets the value of :py:attr:`numPartitions`.
        """
        return self._set(numPartitions=value)

    def getNumPartitions(self):
        """
        Gets the value of :py:attr:`numPartitions`.
        """
        return self.getOrDefault(self.numPartitions)


class HasDevice(Params):
    device = Param(
        Params._dummy(),
        "device",
        "Device.",
        typeConverter=TypeConverters.toInt,
    )

    def setDevice(self, value):
        """
        Sets the value of :py:attr:`device`.
        """
        return self._set(device=value)

    def getDevice(self):
        """
        Gets the value of device or its default value.
        """
        return self.getOrDefault(self.device)


class HasBatchSize(Params):
    batchSize = Param(
        Params._dummy(),
        "batchSize",
        "Batch size.",
        typeConverter=TypeConverters.toInt,
    )

    def setBatchSize(self, value):
        """
        Sets the value of :py:attr:`batchSize`.
        """
        return self._set(batchSize=value)

    def getBatchSize(self):
        """
        Gets the value of batchSize or its default value.
        """
        return self.getOrDefault(self.batchSize)


class HasWhiteList(Params):
    """
    Mixin for param whiteList.
    """

    whiteList: "Param[List[str]]" = Param(
        Params._dummy(),
        "whiteList",
        "White list.",
        typeConverter=TypeConverters.toListString,
    )

    def __init__(self) -> None:
        super(HasWhiteList, self).__init__()

    def getWhiteList(self) -> Any:
        """
        Gets the value of whiteList or its default value.
        """
        return self.getOrDefault(self.whiteList)

    def setWhiteList(self, value: list[str]) -> Any:
        """
        Sets the value of :py:attr:`whiteList`.
        """
        return self._set(whiteList=value)


class HasBlackList(Params):
    """
    Mixin for param blackList.
    """

    blackList: "Param[List[str]]" = Param(
        Params._dummy(),
        "blackList",
        "Black list.",
        typeConverter=TypeConverters.toListString,
    )

    def getBlackList(self) -> list[str]:
        """
        Gets the value of whiteList or its default value.
        """
        return self.getOrDefault(self.blackList)

    def setBlackList(self, value: str) -> Any:
        """
        Sets the value of :py:attr:`blackList`.
        """
        return self._set(blackList=value)


class HasScoreThreshold(Params):
    """
    Mixin for param scoreThreshold.
    """

    scoreThreshold = Param(
        Params._dummy(),
        "scoreThreshold",
        "Score threshold.",
        typeConverter=TypeConverters.toFloat,
    )

    def __init__(self) -> None:
        super(HasScoreThreshold, self).__init__()

    def getScoreThreshold(self) -> float:
        """
        Gets the value of scoreThreshold or its default value.
        """
        return self.getOrDefault(self.scoreThreshold)

    def setScoreThreshold(self, value: float) -> Any:
        """
        Sets the value of :py:attr:`scoreThreshold`.
        """
        return self._set(scoreThreshold=value)


class HasModel(Params):
    """
    Mixin for param model.
    """

    model = Param(
        Params._dummy(),
        "model",
        "Model.",
        typeConverter=TypeConverters.toString,
    )

    def __init__(self) -> None:
        super(HasModel, self).__init__()

    def getModel(self) -> str:
        """
        Gets the value of model or its default value.
        """
        return self.getOrDefault(self.model)

    def setModel(self, value: str) -> Any:
        """
        Sets the value of :py:attr:`model`.
        """
        return self._set(model=value)


class HasColor(Params):
    color = Param(
        Params._dummy(),
        "color",
        "Color.",
        typeConverter=TypeConverters.toString,
    )

    def setColor(self, value: str) -> Any:
        """
        Sets the value of :py:attr:`color`.
        """
        return self._set(color=value)

    def getColor(self) -> str:
        """
        Gets the value of color or its default value.
        """
        return self.getOrDefault(self.color)


class HasDefaultEnum(Params):

    @staticmethod
    def _any_lang_to_code(lang):
        if lang in TESSERACT_CODE_TO_LANGUAGE:
            return LANGUAGE_TO_CODE[TESSERACT_CODE_TO_LANGUAGE[lang]]
        if lang in LANGUAGE_TO_CODE:
            return LANGUAGE_TO_CODE[lang]
        if lang in CODE_TO_LANGUAGE:
            return lang
        raise ValueError(f"Invalid language: {lang}")

    def _setDefault(self, **kwargs: Any) -> Any:
        """
        Sets default params.
        """
        for param, val in kwargs.items():
            value = val
            if param == "lang":
                value = [self._any_lang_to_code(lang) for lang in value]
            if value is not None and isinstance(value, Enum):
                try:
                    value = value.value
                except TypeError as e:
                    raise TypeError(
                        'Invalid default param value given for param "%s". %s'
                        % (param, e),
                    ) from e
            validator = "validate" + param[0].upper() + param[1:]
            if hasattr(self, validator):
                value = getattr(self, validator)(value)
            super(HasDefaultEnum, self)._setDefault(**{param: value})
        return self

    def _set(self, **kwargs: Any) -> Any:
        """
        Sets user-supplied params.
        """
        for param, val in kwargs.items():
            p = getattr(self, param)
            value = val
            if value is not None:
                if p.name == "lang":
                    value = [self._any_lang_to_code(lang) for lang in value]
                if isinstance(value, Enum):
                    value = value.value
                validator = "validate" + param[0].upper() + param[1:]
                if hasattr(self, validator):
                    value = getattr(self, validator)(value)
                try:
                    value = p.typeConverter(value)
                except TypeError as e:
                    raise TypeError(
                        'Invalid param value given for param "%s". %s' % (p.name, e),
                    ) from e
            self._paramMap[p] = value
        return self


class HasColumnValidator:
    """Column validator."""

    def _validate(self, column_name: str, dataset: Any) -> Any:
        """
        Validate input schema.
        """
        if column_name not in dataset.columns:
            raise ValueError(
                f"Missing input column in transformer {self.uid}: "
                f"Column '{column_name}' is not present.",
            )
        return dataset[column_name]


class HasPartitionMap(Params):
    """Has Partition Map."""

    partitionMap = Param(
        Params._dummy(),
        "partitionMap",
        "Force use pandas udf.",
        typeConverter=TypeConverters.toBoolean,
    )

    def setPartitionMap(self, value: bool) -> Any:
        """
        Sets the value of :py:attr:`partitionMap`.
        """
        return self._set(partitionMap=value)

    def getPartitionMap(self) -> bool:
        """
        Sets the value of :py:attr:`partitionMap`.
        """
        return self.getOrDefault(self.partitionMap)


class HasLang(Params):
    """Has Lang."""

    lang = Param(
        Params._dummy(),
        "lang",
        "Language.",
        typeConverter=TypeConverters.toListString,
    )

    def setLang(self, value: str) -> Any:
        """
        Sets the value of :py:attr:`lang`.
        """
        return self._set(lang=value)

    def getLang(self) -> str:
        """
        Gets the value of lang or its default value.
        """
        return self.getOrDefault(self.lang)

    def getLangTess(self) -> str:
        """
        Gets the value of lang or its default value.
        """
        return "+".join(
            LANGUAGE_TO_TESSERACT_CODE[CODE_TO_LANGUAGE[lang]]
            for lang in self.getOrDefault(self.lang)
        )


class HasSchema(Params):
    """Has Schema."""

    schema = Param(
        Params._dummy(),
        "schema",
        "Output schema.",
        typeConverter=TypeConverters.toString,
    )

    schemaByPrompt = Param(
        Params._dummy(),
        "schemaByPrompt",
        "Output schema by prompt.",
        typeConverter=TypeConverters.toBoolean,
    )

    @staticmethod
    def toPydanticSchema(schema: str) -> Any:
        schema = json.loads(schema)
        return json_schema_to_model(schema, schema.get("$defs", {}))

    def getPaydanticSchema(self) -> Any:
        return self.toPydanticSchema(self.getSchema())

    def validateSchema(self, value: Any) -> Any:
        """
        Validate schema.
        """
        if isinstance(value, str):
            return value
        if issubclass(value, BaseModel):
            value = json.dumps(value.model_json_schema())
        return value

    def getSchema(self) -> str:
        """
        Gets the value of schema or its default value.
        """
        return self.getOrDefault(self.schema)

    def setSchema(self, value: bool) -> Any:
        """
        Sets the value of :py:attr:`schema`.
        """
        return self._set(schema=value)

    def getSchemaByPrompt(self) -> bool:
        """
        Gets the value of schemaByPrompt or its default value.
        """
        return self.getOrDefault(self.schemaByPrompt)


class HasPrompt(Params):

    prompt = Param(
        Params._dummy(),
        "prompt",
        "Prompt.",
        typeConverter=TypeConverters.toString,
    )

    def getPrompt(self) -> str:
        """
        Gets the value of prompt or its default value.
        """
        return self.getOrDefault(self.prompt)

    def setPrompt(self, value: str) -> Any:
        """
        Sets the value of :py:attr:`prompt`.
        """
        return self._set(prompt=value)


class HasLLM(Params):
    """
    Mixin for param model.
    """

    model = Param(
        Params._dummy(),
        "model",
        "Model.",
        typeConverter=TypeConverters.toString,
    )
    apiBase = Param(
        Params._dummy(),
        "apiBase",
        "apiBase.",
        typeConverter=TypeConverters.toString,
    )
    apiKey = Param(
        Params._dummy(),
        "apiKey",
        "apiKey.",
        typeConverter=TypeConverters.toString,
    )
    delay = Param(
        Params._dummy(),
        "delay",
        "Delay.",
        typeConverter=TypeConverters.toInt,
    )
    maxRetry = Param(
        Params._dummy(),
        "maxRetry",
        "Max retry.",
        typeConverter=TypeConverters.toInt,
    )
    temperature = Param(
        Params._dummy(),
        "temperature",
        "Temperature.",
        typeConverter=TypeConverters.toFloat,
    )
    systemPrompt = Param(
        Params._dummy(),
        "systemPrompt",
        "System prompt.",
        typeConverter=TypeConverters.toString,
    )

    openAiClient = None

    def __init__(self) -> None:
        super(HasLLM, self).__init__()

    def getOIClient(self) -> Any:
        """OpenAI client."""
        if self.openAiClient:
            return self.openAiClient
        self.openAiClient = self.getClient(self.getApiKey(), self.getApiBase())
        return self.openAiClient

    @classmethod
    def getClient(cls, apiKey: str, apiBase: str) -> Any:
        """OpenAI client."""
        from openai import OpenAI

        kwargs = {}
        if apiKey:
            kwargs["api_key"] = apiKey
        if apiBase:
            kwargs["base_url"] = apiBase
        if not apiKey and not os.environ.get("OPENAI_API_KEY"):
            raise ValueError(
                "One of `apiKey` param or environment variable `OPENAI_API_KEY` must be provided.",
            )
        return OpenAI(**kwargs)

    def getDelay(self) -> int:
        """
        Gets the value of delay or its default value.
        """
        return self.getOrDefault(self.delay)

    def getMaxRetry(self) -> int:
        """
        Gets the value of maxRetry or its default value.
        """
        return self.getOrDefault(self.maxRetry)

    def getModel(self) -> Any:
        """
        Gets the value of model or its default value.
        """
        return self.getOrDefault(self.model)

    def setModel(self, value: str) -> Any:
        """
        Sets the value of :py:attr:`model`.
        """
        return self._set(model=value)

    def getApiBase(self) -> str:
        """
        Gets the value of apiBase or its default value.
        """
        return self.getOrDefault(self.apiBase)

    def setApiBase(self, value: str) -> Any:
        """
        Sets the value of :py:attr:`model`.
        """
        return self._set(apiBase=value)

    def getApiKey(self) -> str:
        """
        Gets the value of apiKey or its default value.
        """
        return self.getOrDefault(self.apiKey)

    def setApiKey(self, value: str) -> Any:
        """
        Sets the value of :py:attr:`apiKey`.
        """
        return self._set(apiKey=value)

    def getSystemPrompt(self) -> str:
        """
        Gets the value of systemPrompt or its default value.
        """
        return self.getOrDefault(self.systemPrompt)

    def setSystemPrompt(self, value: str) -> Any:
        """Sets the value of :py:attr:`systemPrompt`."""
        return self._set(systemPrompt=value)

    def getTemperature(self) -> float:
        """Gets the value of temperature or its default value."""
        return self.getOrDefault(self.temperature)


class HasPropagateExc(Params):
    """Propagate error."""

    propagateError = Param(
        Params._dummy(),
        "propagateError",
        "propagateError.",
        typeConverter=TypeConverters.toBoolean,
    )

    def getPropagateError(self) -> bool:
        """Gets the value of propagateError or its default value."""
        return self.getOrDefault(self.propagateError)

    def setPropagateError(self, value: bool) -> Any:
        """Sets the value of :py:attr:`propagateError`."""
        return self._set(propagateError=value)



import sys

import cv2
import numpy as np


class NormalizeImage(object):
    """normalize image such as substract mean, divide std"""

    def __init__(self, scale=None, mean=None, std=None, order="chw", **kwargs):
        if isinstance(scale, str):
            scale = eval(scale)
        self.scale = np.float32(scale if scale is not None else 1.0 / 255.0)
        mean = mean if mean is not None else [0.485, 0.456, 0.406]
        std = std if std is not None else [0.229, 0.224, 0.225]

        shape = (3, 1, 1) if order == "chw" else (1, 1, 3)
        self.mean = np.array(mean).reshape(shape).astype("float32")
        self.std = np.array(std).reshape(shape).astype("float32")

    def __call__(self, data):
        img = data["image"]
        from PIL import Image

        if isinstance(img, Image.Image):
            img = np.array(img)
        assert isinstance(img, np.ndarray), "invalid input 'img' in NormalizeImage"
        data["image"] = (img.astype("float32") * self.scale - self.mean) / self.std
        return data


class DetResizeForTest(object):
    def __init__(self, **kwargs):
        super(DetResizeForTest, self).__init__()
        self.resize_type = 0
        self.keep_ratio = False
        if "image_shape" in kwargs:
            self.image_shape = kwargs["image_shape"]
            self.resize_type = 1
            if "keep_ratio" in kwargs:
                self.keep_ratio = kwargs["keep_ratio"]
        elif "limit_side_len" in kwargs:
            self.limit_side_len = kwargs["limit_side_len"]
            self.limit_type = kwargs.get("limit_type", "min")
        elif "resize_long" in kwargs:
            self.resize_type = 2
            self.resize_long = kwargs.get("resize_long", 960)
        else:
            self.limit_side_len = 736
            self.limit_type = "min"

    def __call__(self, data):
        img = data["image"]
        src_h, src_w, _ = img.shape
        if sum([src_h, src_w]) < 64:
            img = self.image_padding(img)

        if self.resize_type == 0:
            # img, shape = self.resize_image_type0(img)
            img, [ratio_h, ratio_w] = self.resize_image_type0(img)
        elif self.resize_type == 2:
            img, [ratio_h, ratio_w] = self.resize_image_type2(img)
        else:
            # img, shape = self.resize_image_type1(img)
            img, [ratio_h, ratio_w] = self.resize_image_type1(img)
        data["image"] = img
        data["shape"] = np.array([src_h, src_w, ratio_h, ratio_w])
        return data

    def image_padding(self, im, value=0):
        h, w, c = im.shape
        im_pad = np.zeros((max(32, h), max(32, w), c), np.uint8) + value
        im_pad[:h, :w, :] = im
        return im_pad

    def resize_image_type1(self, img):
        target_h, target_w = self.image_shape
        ori_h, ori_w = img.shape[:2]
        if self.keep_ratio:
            scale = min(target_h / ori_h, target_w / ori_w)
            resize_h = int(ori_h * scale)
            resize_w = int(ori_w * scale)
            img = cv2.resize(img, (resize_w, resize_h))

            if resize_h < target_h:
                pad_bottom = target_h - resize_h
                img = cv2.copyMakeBorder(
                    img, 0, pad_bottom, 0, 0, cv2.BORDER_CONSTANT, value=(255, 255, 255)
                )
            if resize_w < target_w:
                pad_right = target_w - resize_w
                img = cv2.copyMakeBorder(
                    img, 0, 0, 0, pad_right, cv2.BORDER_CONSTANT, value=(255, 255, 255)
                )
            ratio_h = ratio_w = scale
        else:
            img = cv2.resize(img, (target_w, target_h))
            ratio_h = target_h / ori_h
            ratio_w = target_w / ori_w
        return img, [ratio_h, ratio_w]

    def resize_image_type0(self, img):
        """
        resize image to a size multiple of 32 which is required by the network
        args:
            img(array): array with shape [h, w, c]
        return(tuple):
            img, (ratio_h, ratio_w)
        """
        limit_side_len = self.limit_side_len
        h, w, c = img.shape

        # limit the max side
        if self.limit_type == "max":
            if max(h, w) > limit_side_len:
                if h > w:
                    ratio = float(limit_side_len) / h
                else:
                    ratio = float(limit_side_len) / w
            else:
                ratio = 1.0
        elif self.limit_type == "min":
            if min(h, w) < limit_side_len:
                if h < w:
                    ratio = float(limit_side_len) / h
                else:
                    ratio = float(limit_side_len) / w
            else:
                ratio = 1.0
        elif self.limit_type == "resize_long":
            ratio = float(limit_side_len) / max(h, w)
        else:
            raise Exception("not support limit type, image ")
        resize_h = int(h * ratio)
        resize_w = int(w * ratio)

        resize_h = max(int(round(resize_h / 32) * 32), 32)
        resize_w = max(int(round(resize_w / 32) * 32), 32)

        try:
            if int(resize_w) <= 0 or int(resize_h) <= 0:
                return None, (None, None)
            img = cv2.resize(img, (int(resize_w), int(resize_h)))
        except:
            print(img.shape, resize_w, resize_h)
            sys.exit(0)
        ratio_h = resize_h / float(h)
        ratio_w = resize_w / float(w)
        return img, [ratio_h, ratio_w]

    def resize_image_type2(self, img):
        h, w, _ = img.shape

        resize_w = w
        resize_h = h

        if resize_h > resize_w:
            ratio = float(self.resize_long) / resize_h
        else:
            ratio = float(self.resize_long) / resize_w

        resize_h = int(resize_h * ratio)
        resize_w = int(resize_w * ratio)

        max_stride = 128
        resize_h = (resize_h + max_stride - 1) // max_stride * max_stride
        resize_w = (resize_w + max_stride - 1) // max_stride * max_stride
        img = cv2.resize(img, (int(resize_w), int(resize_h)))
        ratio_h = resize_h / float(h)
        ratio_w = resize_w / float(w)

        return img, [ratio_h, ratio_w]


class ToCHWImage(object):
    """convert hwc image to chw image"""

    def __init__(self, **kwargs):
        pass

    def __call__(self, data):
        img = data["image"]
        from PIL import Image

        if isinstance(img, Image.Image):
            img = np.array(img)
        data["image"] = img.transpose((2, 0, 1))
        return data


class KeepKeys(object):
    def __init__(self, keep_keys, **kwargs):
        self.keep_keys = keep_keys

    def __call__(self, data):
        data_list = []
        for key in self.keep_keys:
            data_list.append(data[key])
        return data_list



import gc
import io
from types import MappingProxyType
from typing import Any

from pyspark import keyword_only

from scaledp.models.detectors.BaseDetector import BaseDetector
from scaledp.params import HasBatchSize, HasDevice
from scaledp.schemas.Box import Box
from scaledp.schemas.DetectorOutput import DetectorOutput

from ...enums import Device


class DocTRTextDetector(BaseDetector, HasDevice, HasBatchSize):

    defaultParams = MappingProxyType(
        {
            "inputCol": "image",
            "outputCol": "boxes",
            "keepInputData": False,
            "scaleFactor": 1.0,
            "scoreThreshold": 0.1,
            "device": Device.CPU,
            "batchSize": 2,
            "model": "db_resnet50",
            "partitionMap": False,
            "numPartitions": 0,
            "pageCol": "page",
            "pathCol": "path",
            "propagateError": False,
            "onlyRotated": False,
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(DocTRTextDetector, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    @classmethod
    def call_detector(cls, images, params):
        import torch
        from doctr.file_utils import CLASS_NAME
        from doctr.io import DocumentFile
        from doctr.models import ocr_predictor

        predictor = ocr_predictor(pretrained=True, det_arch=params["model"])

        predictor.det_predictor.model.postprocessor.box_thresh = params[
            "scoreThreshold"
        ]

        detector = predictor.det_predictor

        device = "cpu" if int(params["device"]) == Device.CPU.value else "cuda"

        def resolve_geometry(
            geom: Any,
        ) -> (
            tuple[float, float, float, float]
            | tuple[float, float, float, float, float, float, float, float]
        ):
            if len(geom) == 4:
                return (*geom[0], *geom[1], *geom[2], *geom[3])
            return (*geom[0], *geom[1])

        docs = []
        for image, _image_path in images:
            buff = io.BytesIO()
            image.save(buff, "png")
            docs.extend(DocumentFile.from_images(buff.getvalue()))

        results = detector.to(device)(docs)

        results_final = []
        for result, (image, image_path) in zip(results, images):
            boxes = []
            w, h = image.size
            for geom in result[CLASS_NAME]:
                g = (
                    geom[:-1].tolist()
                    if geom.shape == (5,)
                    else resolve_geometry(geom[:4].tolist())
                )
                boxes.append(
                    Box.from_bbox(
                        [g[0] * w, g[1] * h, g[2] * w, g[3] * h],
                        score=geom[-1],
                    ),
                )
            results_final.append(
                DetectorOutput(path=image_path, type="doctr", bboxes=boxes),
            )

        gc.collect()
        if int(params["device"]) == Device.CUDA.value:
            torch.cuda.empty_cache()

        return results_final



import json
from types import MappingProxyType
from typing import Any, List

import pandas as pd
from pydantic import BaseModel
from pyspark import keyword_only
from pyspark.ml.param import Param, Params, TypeConverters
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_random_exponential,
)

from scaledp.models.ner.BaseNer import BaseNer
from scaledp.params import HasLLM, HasPrompt, HasPropagateExc
from scaledp.schemas.Entity import Entity
from scaledp.schemas.NerOutput import NerOutput

from ...enums import Device


class LLMNer(BaseNer, HasLLM, HasPrompt, HasPropagateExc):

    tags = Param(
        Params._dummy(),
        "tags",
        "Ner tags.",
        typeConverter=TypeConverters.toListString,
    )

    defaultParams = MappingProxyType(
        {
            "inputCols": ["text"],
            "outputCol": "ner",
            "keepInputData": True,
            "whiteList": [],
            "numPartitions": 1,
            "device": Device.CPU,
            "batchSize": 1,
            "scoreThreshold": 0.0,
            "pageCol": "page",
            "pathCol": "path",
            "systemPrompt": "You are excellent NER tag extractor.",
            "prompt": """Please extract text from the image.""",
            "model": "gemini-2.5-flash-lite",
            "apiBase": "",
            "apiKey": "",
            "delay": 30,
            "maxRetry": 6,
            "propagateError": False,
            "tags": [
                "PERSON",
                "LOCATION",
                "DATE",
                "EMAIL",
                "PHONE",
                "ORGANIZATION",
                "ID",
            ],
            "partitionMap": False,
        },
    )

    def getPaydanticSchema(self):

        class Entity(BaseModel):
            entity_group: str
            word: str

        class NerOutput(BaseModel):
            entities: List[Entity]

        return NerOutput

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(LLMNer, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)
        self.pipeline = None

    def transform_udf(self, document, params=None):

        if params is None:
            params = self.get_params()
        params = json.loads(params)

        mapping = []
        for idx, box in enumerate(document.bboxes):
            mapping.extend([idx] * (len(box.text) + 1))

        from openai import RateLimitError

        client = self.getClient(params["apiKey"], params["apiBase"])

        @retry(
            retry=retry_if_exception_type(RateLimitError),
            wait=wait_random_exponential(min=1, max=params["delay"]),
            stop=stop_after_attempt(params["maxRetry"]),
        )
        def completion_with_backoff(**kwargs: Any):
            return client.beta.chat.completions.parse(**kwargs)

        schema = self.getPaydanticSchema().model_json_schema()

        completion = completion_with_backoff(
            model=params["model"],
            messages=[
                {
                    "role": "system",
                    "content": [
                        {"type": "text", "text": params["systemPrompt"]},
                    ],
                },
                {
                    "role": "user",
                    "content": f'Pleas extract NER tags: {",".join(params["tags"])}'
                    f" as json with schema: {schema}. "
                    f"Return only valid json without any other extra text. From the text:"
                    + document.text,
                },
            ],
            response_format={"type": "json_object"},
        )
        entities = []

        data = json.loads(
            completion.choices[0]
            .message.content.replace("```json", "")
            .replace("```", ""),
        )
        for tag in data["entities"]:
            if (
                len(self.getWhiteList()) > 0
                and tag["entity_group"] not in self.getWhiteList()
            ):
                continue
            boxes = []
            word = tag["word"]
            for _idx, box in enumerate(document.bboxes):
                if any(
                    word.lower() in box.text.lower()
                    and (len(word) > 2 or abs(len(word) - len(box.text)) < 2)
                    for word in word.replace(",", " ").replace("@", " ").split(" ")
                    if len(word) > 1
                ):
                    boxes.append(box)
            t = Entity(
                entity_group=tag["entity_group"],
                score=0,  # float(tag["score"]),
                word=tag["word"],
                start=0,  # tag["start"],
                end=0,  # tag["end"],
                boxes=boxes,
            )
            entities.append(t)

        return NerOutput(path=document.path, entities=entities, exception="")

    @staticmethod
    def transform_udf_pandas(
        texts: pd.DataFrame,
        params: pd.Series,
    ) -> pd.DataFrame:  # pragma: no cover
        results = []

        return pd.DataFrame(results)



import logging
import math
import random
import traceback
from types import MappingProxyType
from typing import Any

from PIL import ImageDraw
from pyspark import keyword_only
from pyspark.ml import Transformer
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable
from pyspark.sql.functions import udf

from scaledp.params import (
    AutoParamsMeta,
    HasBlackList,
    HasColor,
    HasColumnValidator,
    HasDefaultEnum,
    HasImageType,
    HasInputCols,
    HasKeepInputData,
    HasNumPartitions,
    HasOutputCol,
    HasPageCol,
    HasWhiteList,
)
from scaledp.schemas.Box import Box
from scaledp.schemas.Entity import Entity
from scaledp.schemas.Image import Image
from scaledp.schemas.NerOutput import NerOutput

from ..enums import ImageType


class ImageDrawBoxes(
    Transformer,
    HasInputCols,
    HasOutputCol,
    HasKeepInputData,
    HasImageType,
    HasPageCol,
    DefaultParamsReadable,
    DefaultParamsWritable,
    HasColor,
    HasNumPartitions,
    HasColumnValidator,
    HasDefaultEnum,
    HasWhiteList,
    HasBlackList,
    metaclass=AutoParamsMeta,
):
    """Draw boxes on image."""

    filled = Param(
        Params._dummy(),
        "filled",
        "Fill rectangle.",
        typeConverter=TypeConverters.toBoolean,
    )

    lineWidth = Param(
        Params._dummy(),
        "lineWidth",
        "Line width.",
        typeConverter=TypeConverters.toInt,
    )

    textSize = Param(
        Params._dummy(),
        "textSize",
        "Text size.",
        typeConverter=TypeConverters.toInt,
    )

    padding = Param(
        Params._dummy(),
        "padding",
        "Padding.",
        typeConverter=TypeConverters.toInt,
    )

    displayDataList = Param(
        Params._dummy(),
        "displayDataList",
        "Display data list.",
        typeConverter=TypeConverters.toListString,
    )

    defaultParams = MappingProxyType(
        {
            "inputCols": ["image", "boxes"],
            "outputCol": "image_with_boxes",
            "keepInputData": False,
            "imageType": ImageType.FILE,
            "filled": False,
            "color": None,
            "lineWidth": 1,
            "textSize": 12,
            "displayDataList": [],
            "numPartitions": 0,
            "padding": 0,
            "pageCol": "page",
            "whiteList": [],
            "blackList": [],
        },
    )

    @keyword_only
    def __init__(self, **kwargs: Any) -> None:
        super(ImageDrawBoxes, self).__init__()
        self._setDefault(**self.defaultParams)
        self._set(**kwargs)

    def getDisplayText(self, box):
        display_list = self.getDisplayDataList()
        text = []
        if display_list:
            for name in display_list:
                if hasattr(box, name):
                    val = getattr(box, name)
                    if isinstance(val, float):
                        text.append(f"{val:0.2f}")
                    elif isinstance(val, int):
                        text.append(str(val))
                    else:
                        text.append(val)
        return ":".join(text)

    def transform_udf(self, image, *data_list: Any):

        def get_color():
            return "#{:06x}".format(random.randint(0, 0xFFFFFF))

        if not isinstance(image, Image):
            image = Image(**image.asDict())
        try:
            if image.exception != "":
                return Image(
                    image.origin,
                    image.imageType,
                    data=bytes(),
                    exception=image.exception,
                )
            img = image.to_pil()
            img1 = ImageDraw.Draw(img)
            fill = self.getColor() if self.getFilled() else None

            for data in data_list:

                if hasattr(data, "entities"):
                    self.draw_ner_boxes(data, fill, get_color, img1)
                else:
                    self.draw_boxes(data, fill, img1)

        except Exception:
            exception = traceback.format_exc()
            exception = f"ImageDrawBoxes: {exception}, {image.exception}"
            logging.warning(exception)
            return Image(image.path, image.imageType, data=bytes(), exception=exception)
        return Image.from_pil(img, image.path, image.imageType, image.resolution)

    def draw_boxes(self, data, fill, img1):
        colors = {}
        for b in data.bboxes:
            box = b
            if not isinstance(box, Box):
                box = Box(**box.asDict())

            # Group by Box.text field for color consistency
            text_key = box.text if hasattr(box, "text") and box.text else "default"

            if text_key not in colors:
                colors[text_key] = "#{:06x}".format(random.randint(0, 0xFFFFFF))

            color = colors[text_key] if self.getColor() is None else self.getColor()

            self.draw_box(box, color, fill, img1)
            text = self.getDisplayText(box)
            if text:
                img1.text(
                    (
                        box.x,
                        box.y - self.getTextSize() * 1.2 - self.getPadding(),
                    ),
                    text,
                    fill=color,
                    font_size=self.getTextSize(),
                )

    def draw_box(self, box, color, fill, img1):
        if box.angle == 0:
            # Draw normal rectangle if angle is 0
            img1.rounded_rectangle(
                box.shape(self.getPadding()),
                outline=color,
                radius=4,
                fill=fill,
                width=self.getLineWidth(),
            )
        else:
            # Draw rotated rectangle for non-zero angles
            center_x = box.x + box.width / 2
            center_y = box.y + box.height / 2
            points = [
                # Top-left
                (-box.width / 2, -box.height / 2),
                # Top-right
                (box.width / 2, -box.height / 2),
                # Bottom-right
                (box.width / 2, box.height / 2),
                # Bottom-left
                (-box.width / 2, box.height / 2),
            ]
            # Rotate points and translate to center
            rotated_points = []

            angle_rad = math.radians(box.angle)
            for px, py in points:
                # Rotate point
                rx = px * math.cos(angle_rad) - py * math.sin(angle_rad)
                ry = px * math.sin(angle_rad) + py * math.cos(angle_rad)
                # Translate to center
                rx += center_x
                ry += center_y
                rotated_points.append((rx, ry))

            # Draw polygon with rotated points
            img1.polygon(
                rotated_points,
                outline=color,
                fill=fill,
                width=self.getLineWidth(),
            )

    def draw_ner_boxes(self, data, fill, get_color, img1):
        black_list = self.getBlackList()
        white_list = self.getWhiteList()
        if not isinstance(data, NerOutput):
            data = NerOutput(**data.asDict())
        colors = {}
        for n in data.entities:
            ner = n
            if not isinstance(ner, Entity):
                ner = Entity(**ner.asDict())

            if ner.entity_group in black_list:
                continue
            if white_list and ner.entity_group not in white_list:
                continue

            if ner.entity_group not in colors:
                colors[ner.entity_group] = get_color()
            color = (
                colors[ner.entity_group] if self.getColor() is None else self.getColor()
            )
            for b in ner.boxes:
                box = b
                if not isinstance(box, Box):
                    box = Box(**box.asDict())
                text = self.getDisplayText(ner)
                self.draw_box(box, color, fill, img1)

                if text:
                    tbox = list(
                        img1.textbbox(
                            (
                                box.x,
                                box.y - self.getTextSize() * 1.2 - self.getPadding(),
                            ),
                            text,
                            font_size=self.getTextSize(),
                        ),
                    )
                    tbox[3] = tbox[3] + self.getTextSize() / 4
                    tbox[2] = tbox[2] + self.getTextSize() / 4
                    tbox[0] = box.x - self.getPadding()
                    tbox[1] = box.y - self.getTextSize() * 1.2 - self.getPadding()
                    img1.rounded_rectangle(
                        tbox,
                        outline=color,
                        radius=2,
                        fill=color,
                    )
                    img1.text(
                        (
                            box.x,
                            box.y - self.getTextSize() * 1.2 - self.getPadding(),
                        ),
                        text,
                        stroke_width=0,
                        fill="white",
                        font_size=self.getTextSize(),
                    )

    def _preprocessing(self, dataset):
        return dataset

    def _transform(self, dataset):
        out_col = self.getOutputCol()
        image_col = self._validate(self.getInputCols()[0], dataset)
        box_cols = [self._validate(col, dataset) for col in self.getInputCols()[1:]]

        dataset = self._preprocessing(dataset)

        if self.getNumPartitions() > 0:
            dataset = dataset.repartition(self.getPageCol()).coalesce(
                self.getNumPartitions(),
            )
        result = dataset.withColumn(
            out_col,
            udf(self.transform_udf, Image.get_schema())(image_col, *box_cols),
        )

        if not self.getKeepInputData():
            result = result.drop(image_col)
        return result
